from collections import OrderedDict
import os
import time

from ...Operation import Operation

inputs=OrderedDict(ppump_controllers=None)
outputs=OrderedDict()
        
class StopPPumps(Operation):
    """Set an array of MitosPPumpControllers to idle."""

    def __init__(self):
        super(StopPPumps,self).__init__(inputs,outputs)
        self.input_doc['ppump_controllers'] = 'list of MitosPPumpController plugins'

    def run(self):
        ppcs = self.inputs['ppump_controllers'] 
        for ipp,ppc in enumerate(ppcs):
            ppc.set_idle()




import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools


# specify workflow names: 
wf_names = ['saxs_fit','read_header','main']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['main']['header_files'] = 'IO.FILESYSTEM.BuildFileList'
op_maps['main']['header_batch'] = 'EXECUTION.Batch'
op_maps['main']['t_filenames'] = 'PACKAGING.BATCH.XYDataFromBatch'
op_maps['main']['t_T'] = 'PACKAGING.BATCH.XYDataFromBatch'
op_maps['main']['saxs_batch'] = 'EXECUTION.Batch'
op_maps['main']['t_params'] = 'PACKAGING.BATCH.XYDataFromBatch'

op_maps['read_header']['read_header'] = 'IO.BL15.ReadHeader_SSRL15'
op_maps['read_header']['time_temp'] = 'PACKAGING.BL15.TimeTempFromHeader'
op_maps['read_header']['saxs_filepath'] = 'IO.FILESYSTEM.BuildFilePath'

op_maps['saxs_fit']['time'] = 'PACKAGING.Identity'
op_maps['saxs_fit']['read_saxs'] = 'IO.NUMPY.Loadtxt_q_I_dI'
op_maps['saxs_fit']['fit_saxs'] = 'PROCESSING.FITTING.XRSDFitGUI'
#op_maps['saxs_fit']['log_I'] = 'PROCESSING.BASIC.LogY'
#op_maps['saxs_fit']['log_Ifit'] = 'PROCESSING.BASIC.LogY'
#op_maps['saxs_fit']['output_CSV'] = 'IO.CSV.WriteArrayCSV'
#op_maps['saxs_fit']['params_path'] = 'IO.FILESYSTEM.BuildFilePath'
#op_maps['saxs_fit']['output_params'] = 'IO.YAML.SaveYAML'

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)

wf = wfmgr.workflows['main']

# input 0: header files location
wf.connect_input('header_dir','header_files.inputs.dir_path') 

# input 1: header file regex
wf.connect_input('header_regex','header_files.inputs.regex') 

wf.connect_output('t_filenames','t_filenames.outputs.x_y')
wf.connect_output('t_params','t_params.outputs.x_y')
wf.connect_output('t_T','t_T.outputs.x_y')

wf.set_op_input('header_batch','work_item','read_header','entire workflow')
wf.set_op_input('header_batch','input_arrays',['header_files.outputs.file_list'],'workflow item')
wf.set_op_input('header_batch','input_keys',['header_filepath'])

wf.set_op_input('t_filenames','batch_outputs','header_batch.outputs.batch_outputs','workflow item')
wf.set_op_input('t_filenames','x_key','time')
wf.set_op_input('t_filenames','y_key','saxs_filepath')
wf.set_op_input('t_filenames','x_sort_flag',True)
wf.set_op_input('t_filenames','x_shift_flag',True)

wf.set_op_input('t_T','batch_outputs','header_batch.outputs.batch_outputs','workflow item')
wf.set_op_input('t_T','x_key','time')
wf.set_op_input('t_T','y_key','temperature')
wf.set_op_input('t_T','x_sort_flag',True)
wf.set_op_input('t_T','x_shift_flag',True)

wf.set_op_input('saxs_batch','work_item','saxs_fit','entire workflow')
wf.set_op_input('saxs_batch','input_arrays',['t_filenames.outputs.x','t_filenames.outputs.y'],'workflow item')
wf.set_op_input('saxs_batch','input_keys',['time','saxs_filepath'])
wf.set_op_input('saxs_batch','pass_thru_params',{'populations':'fit_populations'})

wf.set_op_input('t_params','batch_outputs','saxs_batch.outputs.batch_outputs','workflow item')
wf.set_op_input('t_params','x_key','time')
wf.set_op_input('t_params','y_key','fit_populations')
wf.set_op_input('t_params','x_sort_flag',True)
wf.set_op_input('t_params','x_shift_flag',True)



wf = wfmgr.workflows['read_header']

# input 0: header file
wf.connect_input('header_filepath','read_header.inputs.file_path') 

# inputs 1-2: keys for fetching time and temperature from header dictionary 
wf.connect_input('time_key','time_temp.inputs.temp_key')
wf.connect_input('temp_key','time_temp.inputs.temp_key')

# input 3: directory containing SAXS pattern csv files 
wf.connect_input('saxs_dir','saxs_filepath.inputs.dir_path')

wf.connect_output('time','time_temp.outputs.time')
wf.connect_output('temperature','time_temp.outputs.temperature')
wf.connect_output('saxs_filepath','saxs_filepath.outputs.file_path')
wf.connect_output('saxs_filename','saxs_filepath.outputs.filename')

wf.set_op_input('time_temp','header_dict','read_header.outputs.header_dict','workflow item')
wf.set_op_input('time_temp','time_key','time')
wf.set_op_input('time_temp','temp_key','TEMP')

wf.set_op_input('saxs_filepath','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('saxs_filepath','suffix','_dz_bgsub')
wf.set_op_input('saxs_filepath','ext','csv')



wf = wfmgr.workflows['saxs_fit']

wf.connect_input('time','time.inputs.data')
wf.connect_input('saxs_filepath','read_saxs.inputs.file_path')
wf.connect_input('populations','fit_saxs.inputs.populations')
wf.connect_input('fixed_params','fit_saxs.inputs.fixed_params')
wf.connect_input('bounds','fit_saxs.inputs.param_bounds')
wf.connect_input('constraints','fit_saxs.inputs.param_constraints')
wf.connect_input('source_wavelength','fit_saxs.inputs.source_wavelength')

wf.connect_output('q_I','read_saxs.outputs.q_I')
wf.connect_output('q_I_opt','fit_saxs.outputs.q_I_opt')
wf.connect_output('time','time.outputs.data')
wf.connect_output('fit_populations','fit_saxs.outputs.populations')

wf.set_op_input('read_saxs','delimiter',',')
wf.set_op_input('fit_saxs','q_I','read_saxs.outputs.q_I','workflow item')

#wf.set_op_input('log_I','x_y','read_spectrum.outputs.q_I','workflow item')
#wf.set_op_input('parameterize_spectrum','q_I','read_spectrum.outputs.q_I','workflow item')
#wf.set_op_input('fit_spectrum','populations','parameterize_spectrum.inputs.populations','workflow item')
#wf.set_op_input('log_Ifit','x_y','fit_spectrum.outputs.q_I_opt','workflow item')
#wf.set_op_input('output_CSV','array','fit_spectrum.outputs.q_I_opt','workflow item')
#wf.set_op_input('output_CSV','headers',['q (1/angstrom)','intensity (arb)'])
#wf.set_op_input('output_CSV','dir_path','read_spectrum.outputs.dir_path','workflow item')
#wf.set_op_input('output_CSV','filename','read_spectrum.outputs.filename','workflow item')
#wf.set_op_input('output_CSV','filetag','_fit')
#wf.set_op_input('params_path','dir_path','read_header.outputs.dir_path','workflow item')
#wf.set_op_input('params_path','filename','read_header.outputs.filename','workflow item')
#wf.set_op_input('params_path','suffix','_saxs_params')
#wf.set_op_input('params_path','ext','yaml')
#wf.set_op_input('output_params','file_path','params_path.outputs.file_path','workflow item')
#wf.set_op_input('output_params','data','fit_spectrum.outputs.params','workflow item')

pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','SAXS','BL15','timeseries_gui_fit.wfl'),wfmgr)




from __future__ import print_function
from collections import OrderedDict
from .PawsPlugin import PawsPlugin
import datetime
import time
import sys
import os
from threading import Thread, Condition

import tzlocal
import numpy as np

inputs = OrderedDict(
    dt=1.,
    t_max=None,
    verbose=False)

class Timer(PawsPlugin):
    """A Paws plugin for marking the passage of time"""

    def __init__(self):
        super(Timer,self).__init__(inputs)
        self.input_doc['dt'] = 'Wait time between signals, in seconds'
        self.input_doc['t_max'] = 'Time in seconds after which the timer stops'
        #self.clock = Event()
        # A lock for thread-safe run/stop:
        self.running_lock = Condition()
        # A lock for the "tick" trigger 
        self.dt_lock = Condition()
        self.time_points = []
        self.py_version = int(sys.version[0])
        self.thread_clone = None
        self.proxy = None
        self.t0 = None 
        self.tz = tzlocal.get_localzone()
        self.ep = datetime.datetime.fromtimestamp(0,self.tz)
        self.t0_utc = None 

    def start(self):
        with self.running_lock:
            self.running = True
        self.thread_clone = self.build_clone()
        self.thread_clone.proxy = self
        th = Thread(target=self.thread_clone.run)
        th.start()

    def run(self):
        # self.thread_clone runs this method in its own thread.
        vb = self.inputs['verbose']
        if vb: self.message_callback('timer STARTED')
        with self.proxy.dt_lock:
            dt = self.inputs['dt'] 
            t_max = self.inputs['t_max']
            keep_going = True
            # set zero-time point 
            self.t0 = datetime.datetime.now(self.tz)
            self.t0_utc = (self.t0-self.ep).total_seconds()
            dt_now = self.dt_utc()
            self.proxy.time_points.append(dt_now)
            self.proxy.notify()
            #dt_err = np.mod(dt_now,dt)
            # TODO: implement feedback to hone the timer
        while dt_now < t_max and keep_going:
            if vb: self.message_callback('tick: {}/{}'.format(dt_now,t_max))
            # attempt to nail the next dt point:
            t_rem = dt-np.mod(self.dt_utc(),dt)
            #time.sleep(t_rem-dt_err)
            time.sleep(t_rem)
            # acquire proxy dt lock, take time point, notify, release
            with self.proxy.dt_lock:
                dt_now = self.dt_utc()
                self.proxy.time_points.append(float(dt_now))
                self.proxy.notify()
            #dt_err = np.mod(dt_now,dt)
            # TODO: implement feedback to hone the timer
            # acquire proxy running_lock, check if we are still running
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
                if vb and not keep_going: self.message_callback('timer STOPPED')
        with self.proxy.running_lock:
            if vb: self.message_callback('timer FINISHED')
            self.proxy.stop()

    def notify(self):
        if int(self.py_version) == 2:
            self.dt_lock.notifyAll()
        else:
            self.dt_lock.notify_all()

    def dt_utc(self):
        t = datetime.datetime.now(self.tz)
        return float((t-self.ep).total_seconds()-self.t0_utc)

    def description(self):
        desc = str('Timer Plugin for Paws: '\
            'Signals the passage of time, '\
            'as a convenience for other plugins '\
            'who depend on a clock to trigger their actions')
        return desc

    def get_plugin_content(self):
        return {'time_points':self.time_points}




from collections import OrderedDict

import numpy as np

from ..Operation import Operation

inputs = OrderedDict(x=None,y=None)
outputs = OrderedDict(x_y=None)

class Zip(Operation):
    """Zip two 1d arrays together."""

    def __init__(self):
        super(Zip, self).__init__(inputs, outputs)
        self.input_doc['x'] = '1d array'
        self.input_doc['y'] = '1d array, same size as x'
        self.output_doc['x_y'] = 'n x 2 array containing x and y'

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        x_y = np.array(zip(x, y))
        self.outputs['x_y'] = x_y




from collections import OrderedDict
from functools import partial

import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure
import tkinter
from tkinter import Tk, \
Frame, Canvas, Button, Label, Entry, StringVar, OptionMenu, Scrollbar, Checkbutton
import numpy as np
#from matplotlib import pyplot as plt
#from matplotlib.widgets import Slider, Button, RadioButtons, TextBox
import xrsdkit
from xrsdkit.fitting.xrsd_fitter import XRSDFitter
from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    populations=None,
    fixed_params=None,
    param_bounds=None,
    param_constraints=None,
    source_wavelength=None)
outputs = OrderedDict(
    populations=None,
    report=None,
    q_I_opt=None,
    success_flag=False)
        
class XRSDFitGUI(Operation):
    """Interactively fit a XRSD spectrum."""

    def __init__(self):
        super(XRSDFitGUI, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q(1/Angstrom) versus I(arb).'
        self.input_doc['populations'] = 'dict defining populations, xrsdkit format'
        self.output_doc['populations'] = 'populations with parameters optimized'
        self.output_doc['success_flag'] = 'Boolean indicating whether '\
            'or not the user was satisfied with the fit.'

    def run(self):
        self.q_I = self.inputs['q_I']
        self.populations = self.inputs['populations']
        self.src_wl = self.inputs['source_wavelength']
        self.pf = self.inputs['fixed_params']
        self.pb = self.inputs['param_bounds']
        self.pc = self.inputs['param_constraints']
        self.fit_report = None
        self.q_I_opt = None
        self.success_flag = False
        self.xrsd_fitter = XRSDFitter(self.q_I,self.populations,self.src_wl)

        self.fit_gui = Tk()
        self.fit_gui.title('xrsd profile fitter')
       
        # data structures for maintaining refs to widgets and variables 
        self.pop_frames = OrderedDict()
        self.structure_vars = OrderedDict()
        self.param_frames = OrderedDict()
        self.basis_frames = OrderedDict()
        self.basis_item_frames = OrderedDict()
        self.basis_param_frames = OrderedDict()
        self.fit_report = None
        self.q_I_opt = None

        # create the plots
        self.build_plot_widgets()
        
        # create the widgets for population control
        self.build_entry_widgets()

        # TODO: make these widgets resize when the main window is resized

        # start the tk loop
        self.fit_gui.mainloop()

        # after tk loop exits, finish Operation
        self.finish() 

    def build_plot_widgets(self):
        self.plot_frame = Frame(self.fit_gui,bd=4,relief=tkinter.SUNKEN, background="green")
        self.plot_frame.pack(side=tkinter.LEFT, expand=tkinter.YES,padx=2,pady=2)

        self.fig = Figure(figsize=(8,7))
        self.ax_plot = self.fig.add_subplot(111)
        self.plot_canvas = FigureCanvasTkAgg(self.fig,self.plot_frame)
        self.plot_canvas.get_tk_widget().pack()
        self.draw_plots()

    def build_entry_widgets(self):
        self.scroll_frame = Frame(self.fit_gui)
        self.scroll_frame.pack(side=tkinter.RIGHT,fill='y')
        self.pops_canvas = Canvas(self.scroll_frame, width=450)
        self.scroll_frame.bind_all("<MouseWheel>",  self.on_mousewheel)
        scr = Scrollbar(self.scroll_frame,orient='vertical',command=self.pops_canvas.yview)
        scr.pack(side=tkinter.RIGHT,fill='y')
        self.pops_frame = Frame(self.pops_canvas)
        scroll_canvas_config = lambda ev: self.pops_canvas.configure(scrollregion=self.pops_canvas.bbox("all"))
        self.pops_frame.bind("<Configure>",scroll_canvas_config)
        self.pops_frame.pack(side=tkinter.LEFT,fill='y')
        self.pops_canvas.create_window((0,0),window=self.pops_frame,anchor='nw')
        self.pops_canvas.configure(yscrollcommand=scr.set)
        # TODO: figure out how to get the scroll bar to not overlap the canvas
        # TODO: enable mouse wheel and track pad scrolling 
        #self.pops_canvas.configure(scrollregion=self.pops_canvas.bbox("all"),width=200,height=200)
        #self.pops_canvas.bind("<Button-4>",lambda event: print(event))
        #self.pops_canvas.bind("<Button-5>",lambda event: print(event))
        #self.pops_canvas.bind("<Button-4>", lambda event: event.widget.yview_scroll(-1, UNITS))
        #self.pops_canvas.bind("<Button-5>", lambda event: event.widget.yview_scroll(1, UNITS))
        self.pops_canvas.pack(side=tkinter.LEFT,fill='y')
        self.rebuild_entry_widgets()

    def on_mousewheel(self, event):
        if event.delta == 0:
            event.delta += 100
        self.pops_canvas.yview_scroll(-1 * event.delta, 'units')

    def finish(self):
        # TODO: check whether the user is satisfied with the result, 
        # and include this as an output or as a part of the report
        self.outputs['populations'] = self.populations
        self.outputs['report'] = self.fit_report
        self.outputs['q_I_opt'] = self.q_I_opt
        self.outputs['success_flag'] = self.success_flag

    def draw_plots(self):
        self.ax_plot.clear()
        #n_pops = len(self.populations)
        #if self.ax_plot is not None:
        #    self.fig.delaxes(self.ax_plot)
        #    #self.ax_plot.remove()
        #    self.ax_plot = None
        #self.ax_plot = plt.axes([0.08,0.05*(n_pops+2),0.45,0.9-0.05*(n_pops+2)])
        #self.ax_plot.semilogy(self.q_I[:,0],self.q_I[:,1],lw=2,color='black')
        self.ax_plot.loglog(self.q_I[:,0],self.q_I[:,1],lw=2,color='black')
        I_est = xrsdkit.scattering.compute_intensity(self.q_I[:,0],self.populations,self.src_wl)
        #self.ax_plot.semilogy(self.q_I[:,0],I_est,lw=2,color='red')
        self.ax_plot.loglog(self.q_I[:,0],I_est,lw=2,color='red')
        self.ax_plot.set_xlabel('q (1/Angstrom)')
        self.ax_plot.set_ylabel('Intensity (counts)')
        self.ax_plot.legend(['measured','computed'])
        self.plot_canvas.show()

    def destroy_entry_widgets(self):
        pop_nm_list = list(self.pop_frames.keys())
        for pop_nm in pop_nm_list: 
            popfrm = self.pop_frames.pop(pop_nm)
            popfrm.pack_forget()
            popfrm.destroy()
            self.structure_vars.pop(pop_nm)
            param_nm_list = list(self.param_frames[pop_nm].keys())
            for param_nm in param_nm_list:
                param_frm = self.param_frames[pop_nm].pop(param_nm)
                param_frm.pack_forget()
                param_frm.destroy()
            site_nm_list = list(self.basis_frames[pop_nm].keys())
            for site_nm in site_nm_list: 
                site_frm = self.basis_frames[pop_nm].pop(site_nm)
                site_frm.pack_forget()
                site_frm.destroy()
                specie_nm_list = list(self.basis_item_frames[pop_nm][site_nm].keys())
                for specie_nm in specie_nm_list: 
                    specie_frms = self.basis_item_frames[pop_nm][site_nm][specie_nm]
                    for ispec in range(len(specie_frms))[::-1]:
                        specie_frm = self.basis_item_frames[pop_nm][site_nm][specie_nm].pop(ispec)
                        specie_frm.pack_forget()
                        specie_frm.destroy()
                        bparam_frms = self.basis_param_frames[pop_nm][site_nm][specie_nm][ispec]
                        bparam_nm_list = list(self.basis_param_frames[pop_nm][site_nm][specie_nm][ispec].keys())
                        for bparam_nm in bparam_nm_list:
                            bparam_frm = bparam_frms.pop(bparam_nm)
                            bparam_frm.pack_forget()
                            bparam_frm.destroy()

    def rebuild_entry_widgets(self):
        self.destroy_entry_widgets()
        # create a frame for every population
        for ipop,pop_name in enumerate(self.populations.keys()):
            popd = self.populations[pop_name]
            pf = Frame(self.pops_frame,bd=4,pady=10,padx=10,relief=tkinter.RAISED) 
            self.pop_frames[pop_name] = pf
            self.param_frames[pop_name] = OrderedDict()
            self.basis_frames[pop_name] = OrderedDict()
            self.basis_item_frames[pop_name] = OrderedDict()
            self.basis_param_frames[pop_name] = OrderedDict()
            pf.pack(side=tkinter.TOP,pady=2,padx=2)

            popl = Label(pf,text='population name:',anchor='e')
            popl.grid(row=0,column=0,sticky=tkinter.E)
            popl2 = Label(pf,text=pop_name)
            popl2.grid(row=0,column=1,sticky=tkinter.W)
            #pope = Entry(pf,width=20)
            #pope.insert(0,pop_name)
            #pope.grid(row=0,column=1,sticky=tkinter.W)
            rmb = Button(pf,text='Remove',command=partial(self.remove_population,pop_name))
            rmb.grid(row=0,column=2)

            strl = Label(pf,text='structure:',width=12,anchor='e')
            strl.grid(row=1,column=0,sticky=tkinter.E)
            strvar = StringVar(pf)
            str_option_dict = OrderedDict.fromkeys(xrsdkit.structure_names)
            strcb = OptionMenu(pf,strvar,*str_option_dict)
            strvar.set(popd['structure'])
            strvar.trace('w',partial(self.update_structure,pop_name))
            strcb.grid(row=1,column=1,sticky=tkinter.W)
            self.structure_vars[pop_name] = strvar
            # TODO: connect stcb to structure selection    ????? is it done?

            paramsl = Label(pf,text='------ PARAMETERS ------')
            paramsl.grid(row=2,column=0,columnspan=3)
            npars = len(popd['parameters'])
            for ip,param_name in enumerate(xrsdkit.structure_params[popd['structure']]):
                paramf = Frame(pf,bd=2,pady=4,padx=10,relief=tkinter.GROOVE) 
                self.param_frames[pop_name][param_name] = paramf
                paramf.grid(row=3+ip,column=0,columnspan=4,sticky=tkinter.E+tkinter.W)
                
                pl = Label(paramf,text='{}:'.format(param_name),width=12,anchor='e')
                pl.grid(row=0,column=0,sticky=tkinter.E)
                pe = Entry(paramf,width=20)

                param_val = xrsdkit.param_defaults[param_name]
                if param_name in popd['parameters']:
                    param_val = popd['parameters'][param_name]
                pe.insert(0,str(param_val))
                pe.grid(row=0,column=1,columnspan=2,sticky=tkinter.W)

                psw = Checkbutton(paramf,text="variable")
                varparam = not xrsdkit.fixed_param_defaults[param_name]
                # TODO: check for param in fixed_params 
                if varparam: psw.select()
                psw.grid(row=0,column=3,sticky=tkinter.W)
                pbndl = Label(paramf,text='bounds:',width=10,anchor='e')
                pbndl.grid(row=1,column=0,sticky=tkinter.E)
                pbnde1 = Entry(paramf,width=8) 
                pbnde2 = Entry(paramf,width=8)
                pbnde1.grid(row=1,column=1,sticky=tkinter.W) 
                pbnde2.grid(row=1,column=2,sticky=tkinter.W) 
                lbnd = xrsdkit.param_bound_defaults[param_name][0]
                ubnd = xrsdkit.param_bound_defaults[param_name][1] 
                # TODO: check for param in param_bounds 
                pbnde1.insert(0,str(lbnd))  
                pbnde2.insert(0,str(ubnd))
                pexpl = Label(paramf,text='expression:',width=10,anchor='e')
                pexpl.grid(row=2,column=0,sticky=tkinter.E)
                pexpe = Entry(paramf,width=16)
                # TODO: check for param in param_constraints 
                pexpe.grid(row=2,column=1,columnspan=3,sticky=tkinter.E+tkinter.W) 
                # TODO: connect pe to param entry
                # TODO: connect psw to changing fixed_params
                # TODO: connect pbnde to changing param_bounds
                # TODO: connect pexpe to setting param_constraints

            basisl = Label(pf,text='--------- BASIS ---------')
            basisl.grid(row=3+npars,column=0,columnspan=3)
            for ist,site_name in enumerate(popd['basis'].keys()):
                basisf = Frame(pf,bd=2,pady=4,padx=10,relief=tkinter.GROOVE) 
                self.basis_frames[pop_name][site_name] = basisf
                self.basis_item_frames[pop_name][site_name] = OrderedDict()
                self.basis_param_frames[pop_name][site_name] = OrderedDict()
                basisf.grid(row=4+npars+ist,column=0,columnspan=4,sticky=tkinter.E+tkinter.W)
                site_def = popd['basis'][site_name]
                stl = Label(basisf,text='occupant name:',width=14,anchor='e')
                stl.grid(row=0,column=0,sticky=tkinter.E)
                ste = Entry(basisf,width=20)
                ste.insert(0,site_name)
                ste.grid(row=0,column=1,columnspan=3,sticky=tkinter.W)
                rmb = Button(basisf,text='Remove')
                rmb.grid(row=0,column=4)
                # TODO: connect the entry to renaming the occupant 
                # TODO: connect rmb to deleting the occupant 

                if popd['structure'] in xrsdkit.crystalline_structure_names:
                    coordl = Label(basisf,text='coordinates:',width=12,anchor='e')
                    coorde1 = Entry(basisf,width=6)
                    coorde2 = Entry(basisf,width=6)
                    coorde3 = Entry(basisf,width=6)
                    coordl.grid(row=1,column=0,sticky=tkinter.E)
                    coorde1.grid(row=1,column=1)
                    coorde2.grid(row=1,column=2)
                    coorde3.grid(row=1,column=3)
                    if 'coordinates' in popd['basis'][site_name]:
                        c = popd['basis'][site_name]['coordinates']
                    else:
                        c = [0,0,0]
                    coorde1.insert(0,str(c[0]))
                    coorde2.insert(0,str(c[1]))
                    coorde3.insert(0,str(c[2]))
                    # TODO: connect the entries to setting the coordinates
                    # TODO: controls for varying or constraining coords

                for ispec, specie_name in enumerate(popd['basis'][site_name].keys()):
                    if not specie_name == 'coordinates':
                        specie_def = popd['basis'][site_name][specie_name] 
                        if not isinstance(specie_def,list):
                            specie_def = [specie_def]
                        self.basis_item_frames[pop_name][site_name][specie_name] = []
                        self.basis_param_frames[pop_name][site_name][specie_name] = []
                        for specd in specie_def:
                            bitmf = Frame(basisf,bd=2,padx=10,pady=4,relief=tkinter.GROOVE) 
                            self.basis_item_frames[pop_name][site_name][specie_name].append(bitmf)
                            self.basis_param_frames[pop_name][site_name][specie_name].append(OrderedDict())
                            bitmf.grid(row=2+ispec,column=0,columnspan=5,pady=4,sticky=tkinter.E+tkinter.W)

                            specl = Label(bitmf,text='specie:'.format(ispec),width=12,anchor='e')
                            specl.grid(row=0,column=0,sticky=tkinter.E)

                            specvar = StringVar(basisf)
                            spec_option_dict = OrderedDict.fromkeys(xrsdkit.form_factor_names)
                            speccb = OptionMenu(bitmf,specvar,*spec_option_dict)
                            specvar.set(specie_name)
                            speccb.grid(row=0,column=1,sticky=tkinter.W+tkinter.E)

                            rmspecb = Button(bitmf,text='Remove')
                            rmspecb.grid(row=0,column=2)
                            # TODO: connect speccb to changing the specie
                            # TODO: connect rmspecb to removing the specie

                            for ibp,bparam_name in enumerate(xrsdkit.form_factor_params[specie_name]):
                                bparf = Frame(bitmf,bd=2,padx=4,pady=4,relief=tkinter.GROOVE) 
                                self.basis_param_frames[pop_name][site_name][specie_name][ispec][bparam_name] = bparf
                                bparf.grid(row=1+ibp,column=0,columnspan=3,sticky=tkinter.E+tkinter.W)
                                bparaml = Label(bparf,text='{}:'.format(bparam_name),width=10,anchor='e')
                                bparaml.grid(row=0,column=0,sticky=tkinter.E)
                                bparame = Entry(bparf,width=16)
                                bparam_val = xrsdkit.param_defaults[bparam_name] 
                                if bparam_name in specd: bparam_val = specd[bparam_name] 
                                bparame.insert(0,str(bparam_val))
                                bparame.grid(row=0,column=1,columnspan=2,sticky=tkinter.E+tkinter.W)
                                bparsw = Checkbutton(bparf,text="variable")
                                bparsw.grid(row=0,column=3,sticky=tkinter.W)
                                bparvar = not xrsdkit.fixed_param_defaults[bparam_name]
                                # TODO: check for bparam in fixed_params 
                                if bparvar: bparsw.select()
                                bparbndl = Label(bparf,text='bounds:',width=10,anchor='e')
                                bparbndl.grid(row=1,column=0,sticky=tkinter.E)
                                bparbnde1 = Entry(bparf,width=8) 
                                bparbnde2 = Entry(bparf,width=8)
                                lbnd = xrsdkit.param_bound_defaults[bparam_name][0]
                                ubnd = xrsdkit.param_bound_defaults[bparam_name][1] 
                                # TODO: check for bparam in param_bounds 
                                bparbnde1.insert(0,str(lbnd))  
                                bparbnde2.insert(0,str(ubnd))
                                bparbnde1.grid(row=1,column=1,sticky=tkinter.W) 
                                bparbnde2.grid(row=1,column=2,sticky=tkinter.W) 
                                bparexpl = Label(bparf,text='expression:',width=10,anchor='e')
                                bparexpl.grid(row=2,column=0,sticky=tkinter.E)
                                bparexpe = Entry(bparf,width=16) 
                                # TODO: check for bparam in param_constraints 
                                bparexpe.grid(row=2,column=1,columnspan=3,sticky=tkinter.E+tkinter.W) 
                                # TODO: connect bparame to setting the param
                                # TODO: connect bparsw to changing fixed_params
                                # TODO: connect bparbnde to changing param_bounds
                                # TODO: connect bparexpe to setting param_constraints

                # TODO: frame for adding a new specie

            # TODO: frame for adding a new site

        # TODO: frame for adding a new population
        new_pop = Frame(self.pops_frame,bd=4,pady=10,padx=10,relief=tkinter.RAISED)
        new_pop.pack(side=tkinter.TOP,pady=2,padx=2, fill="both", expand=True)
        add_pop = Label(new_pop,text='add population:',anchor='e')
        add_pop.grid(row=0,column=0,sticky=tkinter.E)
        name = Entry(new_pop,width=20)
        name.insert(0,'unique name')
        name.grid(row=0,column=1,sticky=tkinter.W)
        add_button = Button(new_pop, text='Add', width = 10, command=self.hi)
        add_button.grid(row=0,column=2, sticky=tkinter.E)

        # TODO: frame for fitting controls: 
        #   # fit button, objective readout
        #   # toggles for logI-weighted and error-weighted
        #   # toggle for user satisfaction, button for finishing

        fitting_control = Frame(self.pops_frame,bd=4,pady=10,padx=10,relief=tkinter.RAISED)
        fitting_control.pack(side=tkinter.TOP, pady=2,padx=2, fill="both", expand=True)
        error_weighted_box = Checkbutton(fitting_control,text="error weighted")
        error_weighted = True
        if error_weighted: error_weighted_box.select()
        error_weighted_box.grid(row=0,column=0)
        logI_weighted_box = Checkbutton(fitting_control, text="logI weighted")
        logI_weighted = True
        if logI_weighted: logI_weighted_box.select()
        logI_weighted_box.grid(row=0,column=1)
        #finish_button = Button(finish,text='Finish',command=partial(self.remove_population,pop_name))
        fit_button = Button(fitting_control,text='Fit',width = 10,  command=self.hi)
        fit_button.grid(row=0,column=2)
        obj = Label(fitting_control,text='obj:',anchor='e')
        obj.grid(row=1,column=0,sticky=tkinter.E)
        result = Entry(fitting_control,width=20)
        result.insert(0,'some number')
        result.grid(row=1,column=1,sticky=tkinter.W)

        finish = Frame(self.pops_frame,bd=4,pady=10,padx=10,relief=tkinter.RAISED)
        finish.pack(side=tkinter.TOP, pady=2,padx=2, fill="both", expand=True)
        fit = Checkbutton(finish,text="Good fit")
        if self.success_flag: fit.select()
        fit.grid(row=0,column=0)
        #finish_button = Button(finish,text='Finish',command=partial(self.remove_population,pop_name))
        finish_button = Button(finish,text='Finish',width = 10, command=self.hi)
        finish_button.grid(row=0,column=1)

    def hi(self):
        print('test-test')


    def remove_population(self,pop_name):
        self.populations.pop(pop_name)
        self.draw_plots()
        self.rebuild_entry_widgets()

    def update_structure(self,pop_name,var_name,dummy,mode):
        s = self.structure_vars[pop_name].get() 
        self.populations[pop_name]['structure'] = s
        new_params = OrderedDict.fromkeys(xrsdkit.structure_params[s])
        for pnm in new_params: new_params[pnm] = xrsdkit.param_defaults[pnm]
        self.populations[pop_name]['parameters'] = new_params
        new_settings = OrderedDict.fromkeys(xrsdkit.structure_settings[s])
        for snm in new_settings: new_settings[snm] = xrsdkit.setting_defaults[snm]
        self.populations[pop_name]['settings'] = new_settings
        # if the new structure is crystalline, ensure coordinates are set
        new_basis = self.populations[pop_name]['basis']
        if s in xrsdkit.crystalline_structure_names:
            for site_item_name, site_def in new_basis.items():
                site_def_keys = list(site_def.keys())
                for nm in site_def_keys:
                    if nm in xrsdkit.noncrystalline_ff_names:
                        site_def.pop(nm) 
                if not 'coordinates' in site_def:
                    site_def['coordinates'] = [0.,0.,0.]
        self.draw_plots()
        self.rebuild_entry_widgets()

    #def _rename_population(self,old_name,new_name):
    #    popd = self.populations.pop(old_name)
    #    self.populations[new_name] = popd
    #    self.setup_plots()
    #    self.setup_populations()

    #def _new_population(self,new_pop_name):
    #    self.populations[new_pop_name] = OrderedDict()
    #    self.populations[new_pop_name]['structure'] = 'diffuse' 
    #    self.populations[new_pop_name]['parameters'] = OrderedDict()
    #    self.populations[new_pop_name]['basis'] = OrderedDict()
    #    self.setup_plots()
    #    self.setup_populations()

    def fit(self,event):
        self.message_callback('fitting...')
        self.params,rpt = self.saxs_fitter.fit(self.params)
        self.reset_params()
        self.update_plots()
        self.message_callback('fit complete (obj:{})'.format(rpt['final_objective']))
        #I_opt = saxs_math.compute_saxs(self.q_I[:,0],self.populations,self.opt_params)
        #self.ax_plot.semilogy(self.q_I[:,0],I_opt,lw=2,color='green')
        #self.ax_plot.legend(['measured',
        #    'estimated (obj: {:.2f})'.format(rpt['initial_objective']),
        #    'fit (obj: {:.2f})'.format(rpt['final_objective'])])
        #self.ax_plot.redraw_in_frame()
        #for param_name,axs in self.param_axes.items():
        #    pars = self.opt_params[param_name]
        #    for ax,par in zip(axs,pars):
        #        ax.plot(par,0.,'r*')
        #        ax.redraw_in_frame()




import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools

# specify workflow names: 
wf_names = ['saxs_integrator','batch_saxs_integrator']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['batch_saxs_integrator']['Integrator Setup'] = 'PROCESSING.INTEGRATION.BuildPyFAIIntegrator'
op_maps['batch_saxs_integrator']['Batch Execution'] = 'EXECUTION.Batch'
op_maps['saxs_integrator']['Read Image'] = 'IO.IMAGE.FabIOOpen'
op_maps['saxs_integrator']['Integrate to 2d'] = 'PROCESSING.INTEGRATION.ApplyIntegrator2d'
op_maps['saxs_integrator']['Integrate to 1d'] = 'PROCESSING.INTEGRATION.ApplyIntegrator1d'
op_maps['saxs_integrator']['log(I) 2d'] = 'PROCESSING.BASIC.ArrayLog'
op_maps['saxs_integrator']['log(I) 1d'] = 'PROCESSING.BASIC.LogY'
op_maps['saxs_integrator']['Output CSV'] = 'IO.CSV.WriteArrayCSV'
op_maps['saxs_integrator']['Output Image'] = 'IO.IMAGE.FabIOWrite'

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)


wf = wfmgr.workflows['saxs_integrator']

# input 0: the file path
wf.set_op_input('Read Image','file_path',None)
wf.connect_input('image_path','Read Image.inputs.file_path')

# input 1: the integrator 
wf.set_op_input('Integrate to 1d','integrator',None,'runtime')
wf.set_op_input('Integrate to 2d','integrator',None,'runtime')
wf.connect_input('integrator',
    ['Integrate to 1d.inputs.integrator',
    'Integrate to 2d.inputs.integrator'])

# If we wanted to save some of the outputs:
#wf.connect_output('q_I','Integrate to 1d.outputs.q_I')
#wf.connect_output('chi','Integrate to 2d.outputs.chi')
#wf.connect_output('I_at_q_chi','Integrate to 2d.outputs.I_at_q_chi')

# By default, disable the output operations
wf.disable_op('Output CSV')
wf.disable_op('Output Image')

# the image data is taken from the upstream "Read Image" outputs
wf.set_op_input('Integrate to 1d','image_data','Read Image.outputs.image_data','workflow item')
wf.set_op_input('Integrate to 2d','image_data','Read Image.outputs.image_data','workflow item')

# log(I) operations take data from the corresponding integration outputs
wf.set_op_input('log(I) 1d','x_y','Integrate to 1d.outputs.q_I','workflow item')
wf.set_op_input('log(I) 2d','x','Integrate to 2d.outputs.I_at_q_chi','workflow item')

# data-output operations take data and filepaths from upstream operation outputs
wf.set_op_input('Output CSV','array','Integrate to 1d.outputs.q_I','workflow item')
wf.set_op_input('Output CSV','headers',['q','I'])
wf.set_op_input('Output CSV','dir_path','Read Image.outputs.dir_path','workflow item')
wf.set_op_input('Output CSV','filename','Read Image.outputs.filename','workflow item')
wf.set_op_input('Output CSV','filetag','_q_I')
wf.set_op_input('Output Image','image_data','Integrate to 2d.outputs.I_at_q_chi','workflow item')
# TODO: a more thoughtful header for output images
wf.set_op_input('Output Image','header',{})
wf.set_op_input('Output Image','dir_path','Read Image.outputs.dir_path','workflow item')
wf.set_op_input('Output Image','filename','Read Image.outputs.filename','workflow item')
wf.set_op_input('Output Image','filetag','_integrated')
wf.set_op_input('Output Image','overwrite',True)
wf.set_op_input('Output Image','ext','.edf')



wf = wfmgr.workflows['batch_saxs_integrator']

# input 0: the calibration parameters
# This is read in from Xi-cam at runtime 
wf.set_op_input('Integrator Setup','poni_dict',None)
wf.connect_input('poni_dict','Integrator Setup.inputs.poni_dict')

# input 1: the list of images
# This is read in from Xi-cam at runtime
wf.set_op_input('Batch Execution','input_arrays',None)

wf.set_op_input('Batch Execution','work_item','saxs_integrator','entire workflow')
wf.set_op_input('Batch Execution','input_keys',['image_path'])
wf.set_op_input('Batch Execution','static_inputs',['Integrator Setup.outputs.integrator'],'workflow item')
wf.set_op_input('Batch Execution','static_input_keys',['integrator'])

# save the workflows in a .wfl file
pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','XICAM','batch_saxs_integrator.wfl'),wfmgr)




from __future__ import print_function
from collections import OrderedDict
import copy
from functools import partial
import traceback
import os

from ..models.TreeModel import TreeModel
from ..operations.Operation import Operation
from .. import pawstools

class Workflow(TreeModel):
    """Workflow built from paws Operations, with TreeModel interface.
    
    This and other paws classes are TreeModels
    mostly for graphical considerations,
    where these (pure python) TreeModels can interface 
    with gui-based tree views by adding a relatively thin adapter,
    such as paws.qt.QTreeModel.QTreeModel.
    """

    def __init__(self):
        flag_dict = OrderedDict(select=False,enable=True)
        super(Workflow,self).__init__(flag_dict)
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.operations = self._root_dict
        self.dependencies = {} 
        self.plugin_names = []
        self.message_callback = self.tagged_print
        self.data_callback = self.set_item 
        self.stop_flag = False

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def add_operation(self,op_name,op):
        """Name and add an Operation to the Workflow.

        If `op_name` is not unique, 
        the existing Operation is overwritten.

        Parameters
        ----------
        op_name : str
            name to give to the new Operation 
        """
        #op.message_callback = self.message_callback
        op.data_callback = partial(self.set_op_item,op_name)
        self.set_item(op_name,op)

    def add_plugin(self,plugin_name):
        """Add the `plugin_name` to self.plugin_names"""
        self.plugin_names.append(plugin_name)

    def load_operation(self,op_name,op_setup_dict,op_manager):
        """Load an Operation from a dict that specifies its parameters.

        If `op_name` is not unique, the Operation is overwritten.

        Parameters
        ----------
        op_name : str
            name to be given to the new Operation 
        op_setup_dict : dict
            dict specifying Operation setup
        """
        op_uri = op_setup_dict['op_module']
        if not op_manager.is_op_activated(op_uri):
            op_manager.activate_op(op_uri)
        op = op_manager.get_data_from_uri(op_uri)()
        self.add_operation(op_name,op)
        il_setup_dict = op_setup_dict['inputs']
        for inp_name in op.inputs.keys():
            if inp_name in il_setup_dict.keys():
                tp = il_setup_dict[inp_name]['tp']
                val = il_setup_dict[inp_name]['val']
                self.set_op_input(op_name,inp_name,val,tp)

    def remove_operation(self,op_name):
        """Remove an Operation by providing its name as `op_name`."""
        self.remove_item(op_name)

    def n_operations(self):
        return len(self.operations) 

    def set_op_input(self,op_name,input_name,val,tp=None):
        """Set up the Operation input at `op_input_uri`.

        This changes op.input_locator indicated by `op_input_uri`
        to refer to the new value `val` and type `tp`.
        For basic and runtime input types,
        the value is also loaded directly into op.inputs.
        For runtime inputs, the value is not stored in the input_locator,
        and this is on purpose, to prevent serializing those objects later.

        Parameters
        ----------
        op_name : str
            name of the Operation 
        input_name : str
            name of the input
        val : object
            any object to set as the Operation input value
        tp : str or int, optional
            the input type determines how the Workflow 
            interprets the input value.
            For example, 
            the input can be set to another Operation output,
            to be found at 'process.outputs.out1',
            or simply be set to the string 'process.outputs.out1',
            depending on the specified type.
            Valid types are any string in pawstools.input_types,
            or any integer in pawstools.valid_types.
            If not provided, the type is left at its previous setting.
        """
        if not op_name in self.operations.keys():
            raise KeyError('Operation {} not found in workflow'.format(op_name))
        if not input_name in self.operations[op_name].inputs.keys():
            raise KeyError('Input name {} not valid for Operation {} ({}).'
            .format(input_name,op_name,type(self.operations[op_name]).__name__))
        if tp is not None and not tp in pawstools.valid_types and not tp in pawstools.input_types:
            # tp is neither a string or an enum
            raise ValueError('[{}] failed to parse input type: {}'.format(__name__,tp))
        il = self.operations[op_name].input_locator[input_name]
        if tp is not None:
            if tp in pawstools.input_types:
                tp = pawstools.input_types.index(tp)
            il.tp = pawstools.valid_types[tp]
        if not il.tp == pawstools.runtime_type:
            il.val = val
        else:
            # set inputlocator.val to None, 
            # so that this object will NOT attempt
            # to be serialized
            il.val = None
        if il.tp in [pawstools.basic_type,pawstools.runtime_type]:
            # these types should be loaded for immediate use
            self.set_item(op_name+'.inputs.'+input_name,val)

    def connect_input(self,wf_input_name,op_input_uris):
        if not isinstance(op_input_uris,list):
            op_input_uris = [op_input_uris]
        for u in op_input_uris:
            if not u in self.keys():
                msg = 'Operation input {} not found in {}'\
                .format(u,self.keys())
                raise KeyError(msg)
        self.inputs[wf_input_name] = op_input_uris

    def connect_output(self,wf_output_name,op_output_uris):
        self.outputs[wf_output_name] = op_output_uris

    def break_input(self,wf_input_name):
        self.inputs.pop(wf_input_name)
    
    def break_output(self,wf_output_name):
        self.outputs.pop(wf_output_name)

    def get_outputs(self):
        d = OrderedDict()
        for wf_out_name in self.outputs.keys():
            d[wf_out_name] = self.get_wf_output(wf_out_name)
        return d

    def set_input(self,wf_input_name,val,tp=None):
        """Set a value for all inputs in self.inputs[`wf_input_name`]."""
        urilist = self.inputs[wf_input_name]
        if not isinstance(urilist,list):
            urilist = [urilist]
        for uri in urilist:
            p = uri.split('.')
            if len(p) == 3 and p[1] == 'inputs':
                self.set_op_input(p[0],p[2],val,tp)
            else:
                # TODO: consider whether any other structure should be allowed.
                # For now, raise an exception.
                #self.set_item(uri,val)
                msg = 'uri {} does not seem to '\
                'be an Operation input'.format(uri)
                raise KeyError(msg)

    def get_wf_input_value(self,wf_input_name):
        uri = self.inputs[wf_input_name]
        if isinstance(uri,list):
            uri = uri[0]
        p = uri.split('.')
        il = self.get_data_from_uri(p[0]).input_locator[p[2]]
        return il.val  

    def get_wf_output(self,wf_output_name):
        """Get all outputs in self.outputs[`wf_output_name`]."""
        r = self.outputs[wf_output_name]
        if isinstance(r,list):
            dl = []
            for rr in r:
                if self.contains_uri(rr):
                    dl.append(self.get_data_from_uri(rr))
                else:
                    dl.append(None)
            return dl
        else:
            if self.contains_uri(r):
                return self.get_data_from_uri(r)
            else:
                return None

    def set_dependency(self,op_name,dependency_ops):
        """Set `op_name` to depend on one or more other `dependency_ops`"""
        if not isinstance(dependency_ops,list):
            dependency_ops = [dependency_ops]
        if op_name in self.dependencies.keys():
            self.dependencies[op_name].extend(dependency_ops)
        else:
            self.dependencies[op_name] = dependency_ops

    def execution_stack(self):
        """Determine order of execution and diagnostics for the Workflow.

        Returns a tuple (list,dict). 
        For any Operation that is not ready to execute,
        the dict gives information as to why it is not ready. 
       
        Returns
        -------
        stk : list
            List of lists of Operation names,
            where each list contains Operations
            whose dependencies are satisfied by the Operations above them.
        diag : dict
            Gives diagnostic information for any Operations not ready to run. 
            Keys are operation names, values are diagnostic info (strings).
        """
        stk = []
        valid_wf_inputs = [] 
        diagnostics = {}
        continue_flag = True
        while not self.stack_size(stk) == self.n_operations() and continue_flag:
            ops_rdy = []
            ops_not_rdy = []
            for op_name in self.operations.keys():
                op_rdy = False
                op_diag = {}
                if not self.is_op_enabled(op_name):
                    op_rdy = False
                    op_diag = {op_name:'Operation is disabled'}
                elif not self.stack_contains(op_name,stk):
                    if op_name in self.dependencies.keys():
                        dep_ops = self.dependencies[op_name]
                        if all([nm in valid_wf_inputs for nm in dep_ops]):
                            op_rdy,op_diag = self.is_op_ready(op_name,valid_wf_inputs)
                        else:
                            op_rdy = False
                            op_diag = {op_name:'One or more unsatisfied dependencies ({}) '.format(dep_ops)}
                    else:
                        op_rdy,op_diag = self.is_op_ready(op_name,valid_wf_inputs)
                diagnostics.update(op_diag)
                if op_rdy:
                    ops_rdy.append(op_name)
            if any(ops_rdy):
                stk.append(ops_rdy)
                for op_name in ops_rdy:
                    valid_wf_inputs += self.get_valid_wf_inputs(op_name)
            else:
                continue_flag = False
        return stk,diagnostics

    def is_op_ready(self,op_name,valid_wf_inputs):
        """self.execution_stack() uses this to check if an Operation is ready"""
        op = self.get_data_from_uri(op_name)
        inputs_rdy = []
        diagnostics = {op_name:''} 
        for name,il in op.input_locator.items():
            msg = ''
            if il.tp == pawstools.workflow_item:
                inp_rdy = False
                if isinstance(il.val,list):
                    if all([v in valid_wf_inputs for v in il.val]):
                        inp_rdy = True 
                else:
                    if il.val in valid_wf_inputs: 
                        inp_rdy = True 
                if not inp_rdy:
                    msg = 'Operation input {} (={}) '.format(name,il.val)\
                    + 'not satisfied by valid inputs list: {}'.format(valid_wf_inputs)
            else:
                inp_rdy = True
            inputs_rdy.append(inp_rdy)
            diagnostics[op_name+'.inputs.'+name] = msg
        if all(inputs_rdy):
            op_rdy = True
        else:
            op_rdy = False
        return op_rdy,diagnostics 

    def get_valid_wf_inputs(self,op_name):
        """Get all valid uris referring to Operation data.

        Returns the TreeModel uris of the Operation,
        its inputs and outputs dicts,
        and each of the data items in the inputs and outputs dicts.
        """
        op = self.get_data_from_uri(op_name)
        # valid_wf_inputs should be the operation, its input and output dicts, and their respective entries
        valid_wf_inputs = [op_name,op_name+'.inputs',op_name+'.outputs']
        valid_wf_inputs += [op_name+'.outputs.'+k for k in op.outputs.keys()]
        valid_wf_inputs += [op_name+'.inputs.'+k for k in op.inputs.keys()]
        return valid_wf_inputs


    def run(self):
        """Execute the Workflow.

        All of the operations in the Workflow that are ready
        will be executed in the order obtained from self.execution_stack()
        """
        self.stop_flag = False
        stk,diag = self.execution_stack()
        bad_diag_keys = [k for k in diag.keys() if diag[k]]
        for k in bad_diag_keys:
            self.message_callback('WARNING- {} is not ready: {}'.format(k,diag[k]))
        self.message_callback('workflow queue:'+os.linesep+self.print_stack(stk))
        for lst in stk:
            if self.stop_flag:
                self.message_callback('Workflow stopped.')
                return
            for op_name in lst: 
                self.message_callback('running: {}'.format(op_name))
                op = self.get_data_from_uri(op_name) 
                for inpnm,il in op.input_locator.items():
                    if il.tp == pawstools.workflow_item:
                        self.set_op_item(op_name,'inputs.'+inpnm,self.get_wf_data(il))
                op.stop_flag = False
                op.run() 
                for outnm,outdata in op.outputs.items():
                    #self.set_op_item(op_name,'outputs.'+outnm,outdata)
                    full_uri = op_name+'.outputs.'+outnm
                    self.data_callback(full_uri,outdata)

    def stop(self):
        """Stop the Workflow and all of its Operations."""
        self.stop_flag = True
        stk,diag = self.execution_stack()
        for lst in stk:
            for op_name in lst: 
                op = self.get_data_from_uri(op_name) 
                op.stop()

    def enable_op(self,opname):
        self.set_op_enabled(opname,True)

    def disable_op(self,opname):
        self.set_op_enabled(opname,False)

    def set_op_enabled(self,opname,flag=True):
        op_item = self.get_from_uri(opname)
        op_item.flags['enable'] = flag

    def is_op_enabled(self,opname):
        op_item = self.get_from_uri(opname)
        return op_item.flags['enable']

    def op_enable_flags(self):
        dct = OrderedDict()
        for opnm in self.operations.keys():
            dct[opnm] = self.get_from_uri(opnm).flags['enable']
        return dct

    def setup_dict(self):
        """Return a dict that describes the Workflow setup.""" 
        wf_dict = OrderedDict() 
        for op_name,op in self.operations.items():
            wf_dict[op_name] = op.setup_dict()
        wf_dict['WORKFLOW_INPUTS'] = self.inputs
        wf_dict['WORKFLOW_OUTPUTS'] = self.outputs
        wf_dict['OP_ENABLE_FLAGS'] = self.op_enable_flags()
        return wf_dict

    def build_clone(self):
        """Produce a clone of this Workflow."""
        new_wf = self.clone() 
        new_wf.inputs = copy.deepcopy(self.inputs)
        new_wf.outputs = copy.deepcopy(self.outputs)
        new_wf.dependencies = copy.deepcopy(self.dependencies)
        new_wf.plugin_names = copy.deepcopy(self.plugin_names)
        # NOTE 1: cloned workflows will dump messages to self.message_callback 
        #new_wf.message_callback = self.message_callback
        # NOTE 2: they will also dump their data to self.data_callback.
        # In cases where this is undesirable,
        # e.g. when running multiple clones in parallel, 
        # this data_callback should be disconnected after cloning.
        # This default setting is mostly intended for the use case
        # of cloning the workflow to run it in a separate thread.
        #new_wf.data_callback = self.data_callback
        #new_wf.plugins = OrderedDict()
        #for pgn_name,pgn in self.plugins.items():
        #    new_wf.plugins[pgn_name] = pgn
        for op_name,op in self.operations.items():
            new_op = op.build_clone()
            new_wf.add_operation(op_name,new_op)
            if not self.is_op_enabled(op_name):
                new_wf.disable_op(op_name)
        return new_wf

    @classmethod
    def clone(cls):
        return cls()

    def build_tree(self,x):
        """Return a dict describing a tree-like structure of this object.

        This is a reimplemention of TreeModel.build_tree() 
        to define this object's child tree structure.
        For a Workflow, a dict is provided for each Operation,
        where the operation dict contains the results of calling
        self.build_tree(op.inputs) and self.build_tree(op.outputs). 
        """
        if isinstance(x,Operation):
            d = OrderedDict()
            d['inputs'] = self.build_tree(x.inputs)
            d['outputs'] = self.build_tree(x.outputs)
            return d
        else:
            return super(Workflow,self).build_tree(x) 

    def set_op_item(self,op_name,item_uri,item_data):
        """Subroutine for use with functools.partial for callbacks"""
        full_uri = op_name+'.'+item_uri
        self.set_item(full_uri,item_data)

    def get_wf_data(self,il):
        if isinstance(il.val,list):
            return [self.get_data_from_uri(v) for v in il.val]
        else:
            return self.get_data_from_uri(il.val)

    @staticmethod
    def stack_contains(itm,stk):
        for lst in stk:
            if itm in lst:
                return True
            #for lst_itm in lst:
            #    if isinstance(lst_itm,list):
            #        if stack_contains(itm,lst_itm):
            #            return True
        return False

    @staticmethod
    def stack_size(stk):
        sz = 0
        for lst in stk:
            sz += len(lst)
            #for lst_itm in lst:
            #    if isinstance(lst_itm,list):
            #        sz += stack_size(lst_itm)
            #    else:
            #        sz += 1
        return sz

    @staticmethod
    def print_stack(stk):
        stktxt = ''
        opt_newline = os.linesep
        n_layers = len(stk)
        for i,lst in enumerate(stk):
            if i == n_layers-1:
                opt_newline = ''
            if len(lst) > 1:
                #if isinstance(lst[1],list):
                #    substk = lst[1]
                #    stktxt += ('[\'{}\':\n{}\n]'+opt_newline).format(lst[0],print_stack(lst[1]))
                #else:
                stktxt += ('{}'+opt_newline).format(lst)
            else:
                stktxt += ('{}'+opt_newline).format(lst)
        return stktxt






import numpy as np
from collections import OrderedDict

from xrsdkit.tools.profiler import profile_spectrum

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    dI=None)
outputs = OrderedDict(
    features=None)

class SpectrumProfiler(Operation):
    """Profile a SAXS spectrum and return some numerical metrics.

    This operation profiles a SAXS spectrum (I(q) vs. q)
    by taking various scalar quantities from the data.
    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        super(SpectrumProfiler, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q(1/Angstrom) versus I(arb).'
        self.output_doc['features'] = 'dict profiling the input spectrum. '\
        'See the documentation of saxskit.saxs_fit.profile_spectrum().'

    def run(self):
        q_I = self.inputs['q_I']
        d_prof = profile_spectrum(q_I)
        self.outputs['features'] = d_prof




from collections import OrderedDict
import glob
import os

from ...Operation import Operation
from ... import optools

inputs=OrderedDict(
    dir_path=None,
    regex='*.tif',
    new_files_only=False)
outputs=OrderedDict(file_iterator=None)

class FileIterator(Operation):
    """Iterate over files matching a regex in a directory"""

    def __init__(self):
        super(FileIterator,self).__init__(inputs,outputs)
        self.input_doc['dir_path'] = 'path to directory containing files'
        self.input_doc['regex'] = 'regular expression to select files'
        self.input_doc['new_files_only'] = 'regular expression to select files'

    def run(self):
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        process_existing_files = not self.inputs['new_files_only']
        it = optools.FileSystemIterator(dirpath,rx,process_existing_files) 
        self.outputs['file_iterator'] = it





from __future__ import print_function
from collections import OrderedDict

from pypif import pif
from citrination_client import CitrinationClient as CitClient

from .. import pawstools
from .PawsPlugin import PawsPlugin

inputs = OrderedDict(
    address=None,
    api_key_file=None)

class CitrinationClient(PawsPlugin):
    """PAWS Plugin wrapping a Citrination client"""

    def __init__(self):
        super(CitrinationClient,self).__init__(inputs)
        self.input_doc['address'] = 'web address of citrination instance'
        self.input_doc['api_key_file'] = 'path to a file in the local filesystem containing a valid citrination api key'
        self.client=None

    def start(self):
        self.address = self.inputs['address'] 
        f = open(self.inputs['api_key_file'],'r')
        self.api_key = str(f.readline()).strip()
        f.close()
        self.client = CitClient(self.api_key,self.address)

    def stop(self):
        self.client = None

    def description(self):
        desc = str('Citrination Client Plugin for Paws: '
            + 'This is a container for the Citrination Client module. '
            + 'The Citrination Client connects to a Citrination instance '
            + 'and exposes some parts of the Citrination API. '
            + 'Startup requires the web address of a Citrination instance '
            + 'and an API key that provides access to that instance.')
        return desc





from collections import OrderedDict

from ..Operation import Operation

inputs = OrderedDict(
    dict_keys=None,
    dict_values=None)
outputs = OrderedDict(
    new_dict=None)

class BuildDict(Operation):
    """Create a dict from keys and values"""

    def __init__(self):
        super(BuildDict, self).__init__(inputs, outputs)
        self.input_doc['dict_keys'] = 'list of dict keys'
        self.input_doc['dict_values'] = 'list of dict values'
        self.output_doc['new_dict'] = 'output dict containing keys and values'

    def run(self):
        nd = OrderedDict() 
        for k,v in zip(self.inputs['dict_keys'],self.inputs['dict_values']):
            nd[k] = v
        self.outputs['new_dict'] = nd 




from collections import OrderedDict

from ..Operation import Operation

inputs = OrderedDict(data=None)
outputs = OrderedDict(data=None)

class Identity(Operation):
    """Operation that loads its one input into its one output"""

    def __init__(self):
        super(Identity,self).__init__(inputs,outputs) 
        self.input_doc['data'] = 'this can actually be anything'
        self.output_doc['data'] = 'this ends up being whatever the input was'
        
    def run(self):
        """Load self.inputs['data'] into self.outputs['data']"""
        self.outputs['data'] = self.inputs['data']




from collections import OrderedDict
import os

import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    delimiter=None)
outputs=OrderedDict(
    q=None,
    I=None,
    q_I=None,
    dI=None,
    dir_path=None,
    filename=None) 

class Loadtxt_q_I_dI(Operation):
    """Read in q, I, and (optionally) error estimate dI, with numpy.loadtxt."""

    def __init__(self):
        super(Loadtxt_q_I_dI, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to data file'
        self.input_doc['delimiter'] = 'delimiter used in data file (optional)'
        self.output_doc['filename'] = 'filename with path and extension stripped'
        self.output_doc['q'] = 'array of q values' 
        self.output_doc['I'] = 'array of intensities' 
        self.output_doc['q_I'] = 'n-by-2 array of q and intensity' 
        self.output_doc['dI'] = 'array of intensity error estimates' 
        self.input_datatype['file_path'] = 'str'
        self.input_datatype['delimiter'] = 'str'

    def run(self):
        p = self.inputs['file_path']
        dlm = self.inputs['delimiter']

        self.outputs['dir_path'] = os.path.split(p)[0]
        filename = os.path.split(p)[1]
        filename_noext = os.path.splitext(filename)[0]
        self.outputs['filename'] = filename_noext 

        d = np.loadtxt(p, delimiter=dlm)
        self.outputs['q'] = d[:,0]
        self.outputs['I'] = d[:,1]
        self.outputs['q_I'] = d[:,0:2]
        if d.shape[1] > 2:
            self.outputs['dI'] = d[:,2]





from collections import OrderedDict
import glob
import copy

import numpy as np

from ..Operation import Operation
from .. import optools

inputs=OrderedDict(
    work_item=None,
    inputs=None,
    input_keys=None,
    condition=None,
    run_condition=None)

outputs=OrderedDict(
    outputs=None)
        
# TODO: unify Operation and Workflow APIs to eliminate instance checks

class Conditional(Operation):
    """Conditionally execute a Workflow or Operation"""

    def __init__(self):
        super(Conditional,self).__init__(inputs,outputs)
        self.input_doc['work_item'] = 'the Operation '\
            'or Workflow object to be executed'
        self.input_doc['inputs'] = 'one object for each of '\
            'the `work_item` inputs indicated by `input_keys`.'
        self.input_doc['input_keys'] = 'list of keys for setting '\
            'inputs of the `work_item`.'
        self.input_doc['condition'] = 'condition that determines '\
            'whether or not to run `work_item`.'
        self.input_doc['run_condition'] = '`work_item` is executed '\
            'if `condition` == `run_condition`.'
        self.output_doc['outputs'] = 'if `run_flag` is True, '\
            'this will hold the `work_item` outputs'
        self.input_datatype['input_keys'] = 'list'
        self.input_datatype['inputs'] = 'list'

    def run(self):
        wrk = self.inputs['work_item']
        cond = self.inputs['condition']
        rcond = self.inputs['run_condition']
        inpks = self.inputs['input_keys']
        inpvals = self.inputs['inputs']

        out_dict = wrk.get_outputs()

        if cond == rcond: 
            if any(inpks): 
                for inpk,inpval in zip(inpks,inpvals):
                    wrk.set_input(inpk,inpval)
            wrk.run()
            out_dict = wrk.get_outputs()

        self.outputs['outputs'] = out_dict




import numpy as np
from collections import OrderedDict

from ...Operation import Operation

inputs = OrderedDict(x=None)
outputs = OrderedDict(invx=None)

class InvertBinaryArray(Operation):
    """Swap the zeros and ones of a binary array."""

    def __init__(self):
        super(InvertBinaryArray, self).__init__(inputs, outputs)
        self.input_doc['x'] = 'binary array (ones and zeros, or booleans)'
        self.output_doc['invx'] = 'the binary inverse of x'

    def run(self):
        x = np.array(self.inputs['x'],dtype=bool)
        self.outputs['invx'] = np.array(np.invert(x))




from collections import OrderedDict

import os.path

from ...Operation import Operation

inputs=OrderedDict(
    root_dir_path=None,
    new_dir_name=None)
outputs=OrderedDict(
    new_dir_path=None)

class MakeDirectory(Operation):
    """Create a new directory on the local filesystem"""

    def __init__(self):
        super(MakeDirectory, self).__init__(inputs, outputs)
        self.input_doc['root_dir_path'] = 'filesystem path where the new directory will be rooted'
        self.input_doc['new_dir_name'] = 'name of the new directory'
        self.output_doc['new_dir_path'] = 'path to new directory' 

    def run(self):
        rp = str(self.inputs['root_dir_path'])
        dn = str(self.inputs['new_dir_name'])
        ndp = os.path.join(rp,dn)
        if not os.path.exists(ndp):
            os.mkdir(ndp)        
        self.outputs['new_dir_path'] = ndp




from __future__ import print_function
import abc
import re
from collections import OrderedDict
import copy

from .. import pawstools

class Operation(object):
    """Class template for implementing paws operations"""

    def __init__(self,inputs,outputs):
        self.inputs = OrderedDict(copy.deepcopy(inputs))
        self.outputs = OrderedDict(copy.deepcopy(outputs))
        self.input_locator = OrderedDict.fromkeys(self.inputs.keys())
        self.outputs = OrderedDict.fromkeys(self.outputs.keys()) 
        self.input_doc = OrderedDict.fromkeys(self.inputs.keys()) 

        # input types determine how inputs are handled 
        # when workflows using the operation are executed or serialized
        #self.input_type = OrderedDict.fromkeys(self.inputs.keys()) 

        # input datatypes are used for typecasting,
        # for when values are set indirectly,
        # e.g. through a gui.
        # if an input is not likely to be set via gui,
        # e.g. if it is a complicated or duck-typed object, 
        # this should be left to None,
        # in which case gui applications should have some default behavior.
        self.input_datatype = OrderedDict.fromkeys(self.inputs.keys())

        self.output_doc = OrderedDict.fromkeys(self.outputs.keys()) 
        for name in self.inputs.keys(): 
        #    self.input_type[name] = basic_type 
            self.input_locator[name] = pawstools.InputLocator(pawstools.basic_type,self.inputs[name])
        self.message_callback = self.tagged_print 
        self.data_callback = None 
        self.stop_flag = False

    def __getitem__(self,key):
        if key == 'inputs':
            return self.inputs
        elif key == 'outputs':
            return self.outputs
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def __setitem__(self,key,data):
        if key == 'inputs':
            self.inputs = data
        elif key == 'outputs':
            self.outputs = data
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def keys(self):
        return ['inputs','outputs']

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    #def load_defaults(self):
    #    """
    #    Set default types and values into the Operation.input_locators.
    #    """
    #    for name in self.inputs.keys():
    #        tp = self.input_type[name]
    #        val = self.inputs[name]
    #        self.input_locator[name] = InputLocator(tp,val)

    def run(self):
        """
        Operation.run() should use the Operation.inputs
        and set values for all of the items in Operation.outputs.
        """
        pass

    def stop(self):
        self.stop_flag = True
        for inp_name,il in self.input_locator.items():
            if il.tp == entire_workflow:
                if self.inputs[inp_name]:
                    self.inputs[inp_name].stop()

    def get_outputs(self):
        return self.outputs

    @classmethod
    def clone(cls):
        return cls()

    def build_clone(self):
        """Clone the Operation"""
        new_op = self.clone()
        for nm,il in self.input_locator.items():
            new_il = pawstools.InputLocator()
            new_il.tp = copy.copy(il.tp)
            new_il.val = copy.copy(il.val)
            #if il.tp == entire_workflow:
            #    if self.inputs[nm]:
            #        wf = self.inputs[nm]
            #        new_wf = wf.build_clone()
            #        new_wf.data_callback = wf.data_callback
            #        new_wf.message_callback = wf.message_callback
            #        new_op.inputs[nm] = new_wf
            #elif il.tp == plugin_item:
            #    # plugin items should not be copied:
            #    # they are expected to operate safely in the main thread
            #    new_op.inputs[nm] = self.inputs[nm]
            #else: 
            #    # NOTE: have to implement __deepcopy__
            #    # for whatever the input is.
            #    new_op.inputs[nm] = copy.deepcopy(self.inputs[nm]) 
            new_op.input_locator[nm] = new_il
        #new_op.message_callback = self.message_callback
        #new_op.data_callback = self.data_callback
        return new_op

    def setup_dict(self):
        """Produce a dictionary fully describing the setup of the Operation.

        Returns
        -------
        dct : dict
            Dictionary specifying the module name and input setup 
            for the current state of the Operation
        """
        op_modulename = self.__module__[self.__module__.find('operations'):]
        op_modulename = op_modulename[op_modulename.find('.')+1:]
        dct = OrderedDict() 
        dct['op_module'] = op_modulename
        inp_dct = OrderedDict() 
        for nm,il in self.input_locator.items():
            inp_dct[nm] = {'tp':copy.copy(il.tp),'val':copy.copy(il.val)}
        dct['inputs'] = inp_dct 
        return dct

    def set_input(self,input_name,val):
        self.inputs[input_name] = val    

    def clear_outputs(self):
        for k,v in self.outputs.items():
            self.outputs[k] = None

    def description(self):
        """Provide a string describing the Operation.
        """
        return str(type(self).__name__+": "
        + self.doc_as_string()
        + "\n\n--- Inputs ---"
        + self.input_description() 
        + "\n\n--- Outputs ---"
        + self.output_description())

    def doc_as_string(self):
        if self.__doc__:
            return re.sub("\s\s+"," ",self.__doc__.replace('\n','')) 
        else:
            return "none"

    def input_description(self):
        a = ""
        for name in self.inputs.keys(): 
            if self.input_locator[name]:
                display_val = self.input_locator[name]
            else:
                display_val = self.inputs[name] 
            a = a + '\n\n'+self.parameter_doc(name,display_val,self.input_doc[name])
        return a

    def output_description(self):
        a = ""
        for name,val in self.outputs.items(): 
            a = a + '\n\n'+self.parameter_doc(name,val,self.output_doc[name])
        return a
    
    @staticmethod
    def parameter_doc(name,value,doc):
        if isinstance(value, pawstools.InputLocator):
            tp_str = pawstools.input_types[value.tp]
            v_str = str(value.val)
            return "- name: {} \n- value: {} ({}) \n- doc: {}".format(name,v_str,tp_str,doc) 
        else:
            v_str = str(value)
            return "- name: {} \n- value: {} \n- doc: {}".format(name,v_str,doc) 
                



from collections import OrderedDict
import os

from pypif import pif
import pypif.obj as pifobj 
from citrination_client import PifSystemReturningQuery, PifSystemQuery, DataQuery, DatasetQuery, IdQuery, FieldQuery, Filter

from ...Operation import Operation

inputs=OrderedDict(
    client_plugin=None,
    dsid=None,
    experiment_id=None)
outputs=OrderedDict(pif_list=None)
        
class GetPIFs(Operation):
    """Fetch a list of PIF objects from Citrination"""

    def __init__(self):
        super(GetPIFs,self).__init__(inputs,outputs)
        self.input_doc['client_plugin'] = 'A running CitrinationClient(PawsPlugin)' 
        self.input_doc['dsid'] = 'Citrination data set ID'
        self.input_doc['experiment_id'] = 'EXPERIMENT_ID tag (optional)'
        self.output_doc['pif_list'] = 'List of PIF objects'

    def run(self):
        cl_pgn = self.inputs['client_plugin'] 
        dsid = self.inputs['dsid'] 
        expt_id = self.inputs['experiment_id']

        if expt_id is not None:
            query = self.dsid_query_with_expt_id(dsid,expt_id)
        else:
            query = self.dsid_query(dsid)

        all_hits = []
        n_hits = 0
        self.message_callback('Querying Citrination for records.')
        #import pdb; pdb.set_trace()
        #
        current_result = cl_pgn.client.search(query)
        # TODO: is it wise to use the clone in this way?
        #
        while current_result.hits is not None:
            all_hits.extend(current_result.hits)
            n_current_hits = len(current_result.hits)
            n_hits += n_current_hits
            query.from_index += n_current_hits 
            #self.message_callback('{} found ... '.format(n_hits))
            current_result = cl_pgn.client.search(query)
        self.message_callback('Found {} records.'.format(n_hits))

        pifs = [x.system for x in all_hits]
        self.outputs['pif_list'] = pifs        

    def dsid_query_with_expt_id(self,dsid,expt_id):
        query = PifSystemReturningQuery(
            from_index=0,
            size=100,
            query=DataQuery(
                dataset=DatasetQuery(
                    id=Filter(
                        equal=dsid)),    
                system=PifSystemQuery(
                    ids=IdQuery(
                        name=FieldQuery(
                            filter=Filter(
                                equal='EXPERIMENT_ID')),
                        value=FieldQuery(
                            filter=Filter(
                                equal=expt_id))))))
        return query

    def dsid_query(self,dsid):
        query = PifSystemReturningQuery(
            from_index=0,
            size=100,
            query=DataQuery(
                dataset=DatasetQuery(
                    id=Filter(
                        equal=dsid))))
        return query





from __future__ import print_function
from collections import OrderedDict
import glob
import copy
import os
import sys
from threading import Thread, Condition
if int(sys.version[0]) == 2:
    import Queue as queue
else:
    import queue 

import paramiko

from .PawsPlugin import PawsPlugin
from .. import pawstools

inputs = OrderedDict(
    username=None,
    hostname=None,
    password=None,
    port=22,
    private_key_file=None,
    timer=None,
    verbose=False)

class SSHFileCopier(PawsPlugin):
    """SSH Client for synchronizing a local directory to a remote one."""

    def __init__(self):
        super(SSHFileCopier,self).__init__(inputs)
        self.input_doc['username'] = 'remote user name'
        self.input_doc['hostname'] = 'remote host name'
        self.input_doc['password'] = 'user password on remote host'
        self.input_doc['port'] = 'port for ssh connection, default 22'
        self.input_doc['private_key_file'] = 'path to private RSA key file'
        self.input_doc['timer'] = 'Timer plugin to trigger activities'
        self.input_doc['verbose'] = 'If True, client prints activities'
        #self.input_doc['source_dir'] = 'path to remote directory to monitor'
        #self.input_doc['dest_dir'] = 'path to local directory to receive new files'
        #self.input_doc['sync_old_files'] = 'If true, copy any files '\
        #    'that are found in `source_dir` and not found in `dest_dir`'
        #self.input_doc['copy_new_files'] = 'If true, copy any files '\
        #    'that arrive in `source_dir` after the client is started'
        self.py_version = int(sys.version[0])
        self.running_lock = Condition()
        self.directory_lock = Condition()
        self.remote_dir = []
        self.local_dir = []
        self.commands = queue.Queue() 
        self.sshcl = None
        self.sftpcl = None
        self.thread_clone = None
        self.proxy = None

    def start(self):
        with self.running_lock:
            self.running = True
        self.thread_clone = self.build_clone()
        self.thread_clone.proxy = self
        th = Thread(target=self.thread_clone.run)
        th.start()
        # wait for notification that the client is connected
        with self.thread_clone.running_lock:
            self.thread_clone.running_lock.wait()

    def run(self):
        # this method is run by self.thread_clone 
        username = self.inputs['username']
        hostname = self.inputs['hostname']
        pw = self.inputs['password']
        port = self.inputs['port']
        pkey_file = self.inputs['private_key_file']
        tmr = self.inputs['timer']
        vb = self.inputs['verbose']

        self.sshcl = paramiko.SSHClient()
        self.sshcl.load_system_host_keys()
        if pkey_file is not None:
            k = paramiko.RSAKey.from_private_key_file(pkey_file)
            self.sshcl.connect(hostname, port, pkey=k)
        else:
            self.sshcl.connect(hostname, port, username, pw)
        if vb: self.message_callback('SSH client connected to {} port {}'.format(hostname,port))
        with self.running_lock:
            self.run_notify()
        self.sftpcl = paramiko.SFTPClient.from_transport(self.sshcl.get_transport())

        # make initial file list
        #stdin, stdout, stderr = self.sshcl.exec_command('ls {}'.format(src_dir))
        #rdir_list = stdout.read().strip().split(os.linesep.encode())
        #ldir_list = glob.glob(os.path.join(dest_dir,'*'))
        #with self.directory_lock:
        #    self.remote_dir.extend(rdir_list)
        #    self.local_dir.extend(ldir_list)
        #with self.proxy.directory_lock:
        #    self.proxy.remote_dir.extend(copy.deepcopy(rdir_list))
        #    self.proxy.local_dir.extend(copy.deepcopy(ldir_list))

        # continue to check for new files until stopped
        keep_going = True
        while keep_going:
            with self.proxy.inputs['timer'].dt_lock:
                self.proxy.inputs['timer'].dt_lock.wait()
            #if vb: self.message_callback('checking for commands...')
            while self.commands.qsize() > 0: 
                cmd = self.commands.get()
                #if vb: self.message_callback('found command: {}'.format(cmd))
                self.commands.task_done()
                if cmd[0] == 'copy':
                    self.copy_files(cmd[1],cmd[2],cmd[3])
        #    if vb: self.message_callback('checking for new files...')
        #    self.get_new_files()
        #    
            with self.proxy.inputs['timer'].running_lock:
                if not self.proxy.inputs['timer'].running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        self.sshcl.close()

    def request_copy(self,remote_dir,regex,local_dir):
        #if self.inputs['verbose']: self.message_callback('requesting copy...')
        copy_cmd = ['copy',remote_dir,regex,local_dir]
        self.thread_clone.commands.put(copy_cmd)

    def copy_files(self,remote_dir,regex,local_dir):
        vb = self.inputs['verbose']
        rrx = os.path.join(remote_dir,regex)
        if vb: self.message_callback('attempting copy from {}'.format(rrx))
        stdin, stdout, stderr = self.sshcl.exec_command('ls {}'.format(rrx))
        rdir_list = stdout.read().strip().split(os.linesep.encode())
        ldir_list = glob.glob(os.path.join(local_dir,'*'))
        ldir_file_list = [os.path.split(fpath)[1] for fpath in ldir_list]
        #if vb: self.message_callback('current files: {}'.format(rdir_list))
        #print(rdir_list)
        #with self.directory_lock:
        for fpath in rdir_list:
            fn = os.path.split(fpath.decode())[1] 
            if fn in ldir_file_list:
                pass
                #if vb: self.message_callback('file exists locally: skipping {}'.format(fn))
            else:
                #self.remote_dir.append(fn) 
                #with self.proxy.directory_lock:
                #    # update the proxy's directory
                #    self.proxy.remote_dir.append(fn)
                if vb: self.message_callback('copying new file: {}'.format(fn))
                # use sftpcl to move the file over
                #if copy_new: 
                self.sftpcl.get(os.path.join(remote_dir,fn),os.path.join(local_dir,fn))

    def run_notify(self):
        if int(self.py_version) == 2:
            self.running_lock.notifyAll()
        else:
            self.running_lock.notify_all()

    #def get_plugin_content(self):
    #    with self.directory_lock:
    #        ldir = copy.deepcopy(self.local_dir)
    #        rdir = copy.deepcopy(self.remote_dir)
    #    return {'remote_dir':rdir,'local_dir':ldir}





from collections import OrderedDict

import numpy as np
import pypif.obj as pifobj
from xrsdkit.tools import piftools

from ...Operation import Operation

inputs=OrderedDict(
    recipe=None,
    experiment_id=None,
    t_T=None,
    t_params=None,
    t_populations=None,
    t_reports=None)
outputs=OrderedDict(pif=None)

class PackNPSynthPIF(Operation):
    """
    Analyze a series of PIFs generated in a nanoparticle synthesis experiment
    and produce a PIF that describes the overall experiment. 
    """

    def __init__(self):
        super(PackNPSynthPIF,self).__init__(inputs,outputs)
        self.input_doc['recipe'] = 'dict describing the synthesis recipe'
        self.input_doc['experiment_id'] = 'experiment id'
        self.input_doc['t_T'] = 'n-by-2 array, Temperature vs. time'
        self.input_doc['t_params'] = 'n-by-2 array, params vs. time'
        self.input_doc['t_populations'] = 'n-by-2 array, populations vs. time'
        self.input_doc['t_reports'] = 'n-by-2 array, fit reports vs. time'
        self.output_doc['pif'] = 'pypif.obj.ChemicalSystem object '\
            'describing the synthesis experiment'

    def run(self):
        rcp = self.inputs['recipe']
        expt_id = self.inputs['experiment_id']
        t_T = self.inputs['t_T']
        t_pars = self.inputs['t_params']
        t_pops = self.inputs['t_populations']
        t_rpts = self.inputs['t_reports']

        csys = pifobj.ChemicalSystem()
        csys.uid = expt_id 
        csys.ids = []
        csys.ids.append(piftools.id_tag('EXPERIMENT_ID',expt_id))
        
        if rcp:
            prep,comp = self.process_np_synth_recipe(rcp)
            csys.preparation = prep
            csys.composition = comp

        csys.properties = []

        if t_T is not None:
            csys.properties.append(
                self.property_vs_time(
                    'temperature',t_T,'seconds','degrees C'))

        #if t_pars is not None:
        #    for pname in saxs_math.parameter_keys:
        #        nt = len(t_pars)
        #        # collect the time and parameter list for all times where pname is reported
        #        t = np.array([t_pars[it][0] for it in range(nt) if pname in t_pars[it][1].keys()])
        #        par = [t_pars[it][1][pname] for it in range(nt) if pname in t_pars[it][1].keys()]
        #        # count the instances of this parameter at each time point
        #        npar = np.array([len(pari) for pari in par])
        #        # TODO: this needs some additional work 
        #        # to ensure the continuity of parameter identities,
        #        # e.g., so that a given peak retains its identity
        #        # throughout the synthesis reaction
        #        for ipar in range(max(npar)):
        #            # get all t points at which this parameter exists
        #            tt = t[(npar>ipar)]
        #            # harvest the parameter value at all of those t points
        #            nt_pari = len(tt)
        #            pari = [par[it][ipar] for it in range(nt_pari)]
        #            t_pari = zip(list(tt),pari)
        #            if any(t_pari):
        #                property_name = piftools.parameter_description[pname]
        #                if max(npar) > 1:
        #                    property_name = property_name + ' ({})'.format(ipar)
        #                csys.properties.append(self.property_vs_time(
        #                property_name,t_pari,'seconds',\
        #                piftools.parameter_units[pname]))

        if t_rpts is not None:
            # fit obj vs. time
            nrpts = len(t_rpts)
            t_obj = [[t_rpts[it][0],t_rpts[it][1]['objective_value']] for
                it in range(nrpts) if 'objective_value' in t_rpts[it][1]]
            if any(t_obj):
                csys.properties.append(self.property_vs_time(
                    'spectrum fitting error',t_obj,'arb'))

        self.outputs['pif'] = csys

    def property_vs_time(self,name,t_param,t_units=None,param_units=None):
        pf = pifobj.Property()
        pf.name = name 
        npts = len(t_param) 
        pf.scalars = [pifobj.Scalar(t_param[i][1]) for i in range(npts)]
        if param_units:
            pf.units = param_units 
        pf.conditions = []
        pf.conditions.append( pifobj.Value('reaction time',
            [pifobj.Scalar(t_param[i][0]) for i in range(npts)],
            None,None,None,t_units) )
        return pf 

    def process_np_synth_recipe(self,rcp):
        return None,None




from collections import OrderedDict
from datetime import datetime as dt
import importlib
import os
import re

import yaml

from . import operations
from . import workflows

# Enumeration of valid types
basic_type = 0          # input is specified directly, 
                        # and is of a basic type
                        # that is easy to copy or serialize 

workflow_item = 1       # the address (TreeModel uri)
                        # of an item in the workflow 

entire_workflow = 2     # the name of a Workflow
                        # in the current WfManager

plugin_item = 3         # the address (TreeModel uri)
                        # of an item in the PluginManager

runtime_type = 4        # input is generated and set during execution,
                        # and serialization should not be attempted

valid_types = [basic_type,workflow_item,entire_workflow,plugin_item,runtime_type]
input_types = ['basic','workflow item','entire workflow','plugin item','runtime']
input_datatypes = ['int','float','str','bool','list','dict']

class InputLocator(object):
    """
    Objects of this class are used as containers for inputs to an Operation.
    They contain the information needed to find the relevant input data.
    """
    def __init__(self,tp=basic_type,val=None):
        self.tp = tp
        self.val = val 
        #self.data = None 

p = os.path.abspath(__file__)
# p = (pawsroot)/paws/core/pawstools.py

d = os.path.dirname(p)
# d = (pawsroot)/paws/core/

d = os.path.dirname(d)
# d = (pawsroot)/paws/
sourcedir = str(d)

d = os.path.dirname(d)
# d = (pawsroot)/
rootdir = str(d)

user_homedir = os.path.expanduser("~")

paws_scratch_dir = os.path.join(user_homedir,'.paws_scratch')
paws_cfg_dir = os.path.join(user_homedir,'.paws_cfg')
if not os.path.exists(paws_cfg_dir):
    os.mkdir(paws_cfg_dir)
if not os.path.exists(paws_scratch_dir):
    os.mkdir(paws_scratch_dir)

# Get the code version from the config.py file.
# Reference version string as __version__
with open(os.path.join(sourcedir,'config.py')) as f: 
    exec(f.read())

class WorkflowAborted(Exception):
    pass

class OperationDisabledError(Exception):
    pass

class WfNameError(Exception):
    pass

class PluginNameError(Exception):
    pass

class PluginLoadError(Exception):
    pass

class OperationLoadError(Exception):
    pass

def dtstr():
    """Return date and time as a string"""
    return dt.strftime(dt.now(),'%Y %m %d, %H:%M:%S')

def timestr():
    """Return time as a string"""
    return dt.strftime(dt.now(),'%H:%M:%S')

def save_file(filename,d):
    """
    Create or replace file indicated by filename,
    as a yaml serialization of dict d.
    """
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()
    
def update_file(filename,d):
    """
    Save the items in dict d into filename,
    without removing members not included in d.
    """
    if os.path.exists(filename):
        f_old = open(filename,'r')
        d_old = yaml.load(f_old)
        f_old.close()
        d_old.update(d)
        d = d_old
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()

def save_cfg(cfg_data,cfg_file):
    cfg = open(cfg_file,'w')
    yaml.dump(cfg_data,cfg)
    cfg.close()

def load_cfg(cfg_file):
    cfg = open(cfg_file,'r')
    cfg_data = yaml.load(cfg)
    cfg.close()
    if not cfg_data:
        cfg_data = OrderedDict() 
    return cfg_data



def save_to_dict(wf_manager):
    d = {} 
    d['OP_ACTIVATION_FLAGS'] = {k:True for k in wf_manager.op_manager.keys() if wf_manager.op_manager.is_op_activated(k)}
    d['PAWS_VERSION'] = __version__
    wfman_dict = OrderedDict()
    for wfname,wf in wf_manager.workflows.items():
        wfman_dict[wfname] = wf.setup_dict() 
    d['WORKFLOWS'] = wfman_dict
    pgin_dict = OrderedDict() 
    for pgin_name in wf_manager.plugin_manager.plugins.keys():
        pgin = wf_manager.plugin_manager.get_data_from_uri(pgin_name)
        pgin_dict[pgin_name] = pgin.setup_dict()
    d['PLUGINS'] = pgin_dict
    return d

def save_to_wfl(wfl_filename,wf_manager):
    """Save workflows, plugins, and active operations to a .wfl file.

    The .wfl file is really just a YAML file. 

    Parameters
    ----------
    wfl_filename : str
        full path of the .wfl file to be saved-
        extension is optional, 
        and an existing file will be overwritten.
    """
    if not os.path.splitext(wfl_filename)[1] == '.wfl':
        wfl_filename = wfl_filename + '.wfl'
    d = save_to_dict(wf_manager)
    save_file(wfl_filename,d)

def load_packaged_wfl(workflow_uri,wf_manager):
    wf_mod = importlib.import_module('.'+workflow_uri,workflows.__name__)
    wfl_path = sourcedir
    wfl_path = os.path.join(wfl_path,'core','workflows')
    p = workflow_uri.split('.')
    for mp in p:
        wfl_path = os.path.join(wfl_path,mp)
    wfl_filename = wfl_path+'.wfl'
    load_wfl(wfl_filename,wf_manager)

def load_wfl(wfl_filename,wf_manager):
    """Set up a OpManager, PluginManager, and WfManager from a .wfl file.

    Parameters
    ----------
    wfl_filename : str
        path to a .wfl file to be loaded
    """
    f = open(wfl_filename,'r')
    d = yaml.load(f)
    f.close()
    if 'PAWS_VERSION' in d.keys():
        wfl_version = d['PAWS_VERSION']
    else:
        wfl_version = '0.0.0'
    wfl_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',wfl_version)
    wfl_vparts = list(map(int,wfl_vparts.groups()))
    current_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',__version__)  
    current_vparts = list(map(int,current_vparts.groups()))
    if wfl_vparts[0] < current_vparts[0] or wfl_vparts[1] < current_vparts[1]:
        warnings.warn('WARNING: paws (version {}) '\
        'is trying to load a state built in version {} - '\
        'this is likely to cause things to crash, '\
        'until the workflows and plugins are reviewed/refactored '\
        'under the current version.'.format(__version__,wfl_version))  
    if 'OP_ACTIVATION_FLAGS' in d.keys():
        for op_module,flag in d['OP_ACTIVATION_FLAGS'].items():
            if op_module in operations.op_modules:
                if flag:
                    wf_manager.op_manager.activate_op(op_module)
    if 'WORKFLOWS' in d.keys():
        wf_dict = d['WORKFLOWS']
        for wf_name,wf_setup_dict in wf_dict.items():
            wf_manager.load_workflow(wf_name,wf_setup_dict)
    if 'PLUGINS' in d.keys():
        plugins_dict = d['PLUGINS']
        for plugin_name,plugin_setup_dict in plugins_dict.items():
            wf_manager.plugin_manager.load_plugin(plugin_name,plugin_setup_dict)






from __future__ import print_function
from collections import OrderedDict

import numpy as np
from xrsdkit.fitting.xrsd_fitter import XRSDFitter
from xrsdkit.scattering import compute_intensity

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    populations=None,
    fixed_params=None,
    param_bounds=None,
    param_constraints=None,
    error_weighted=True,
    logI_weighted=True,
    logq_weighted=False,
    source_wavelength=None)

outputs = OrderedDict(
    populations=None,
    report=None,
    q_I_opt=None)

class XRSDFit(Operation):
    """Fit scattering equation parameters to measured data.

    Optimizes the parameters of one or several populations of scatterers
    against a measured intensity profile (I(q) vs. q).
    """

    def __init__(self):
        super(XRSDFit, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of wave vectors (1/Angstrom) and intensities (counts)'
        self.input_doc['populations'] = 'dict specifying '\
            'scatterer populations and initial guesses for parameters, '\
            'formatted according to the xrsdkit specifications'
        self.input_doc['fixed_params'] = 'dict indicating which parameters '\
            'should be held fixed during optimization, '\
            'where the dict structure mirrors `populations`, '\
            'but with boolean values'
        self.input_doc['param_bounds'] = 'like `fixed_params`, '\
            'but with tuple values specifying '\
            'the lower and upper parameter bounds'
        self.input_doc['param_constraints'] = 'like `fixed_params`, '\
            'but with expressions (strings) as values, '\
            'where the expressions define the constraints '\
            'according to the xrsdkit specifications'
        self.input_doc['source_wavelength'] = 'wavelength '\
            'of x-ray source, in Angstroms'

        self.output_doc['populations'] = 'similar to input `populations`, '\
            'but with the optimized parameters'
        self.output_doc['report'] = 'dict expressing '\
            'the numerical results of the optimization'
        self.output_doc['q_I_opt'] = 'n-by-2 array '\
            'of q and the optimized computed intensity profile'
        self.input_datatype['populations'] = 'dict'
        self.input_datatype['fixed_params'] = 'dict'
        self.input_datatype['param_bounds'] = 'dict'
        self.input_datatype['param_constraints'] = 'dict'

    def run(self):
        q_I = self.inputs['q_I']
        pops = self.inputs['populations']
        p_fix = self.inputs['fixed_params']
        p_b = self.inputs['param_bounds']
        p_c = self.inputs['param_constraints']
        errwtd = self.inputs['error_weighted']
        logIwtd = self.inputs['logI_weighted']
        logqwtd = self.inputs['logq_weighted']
        src_wl = self.inputs['source_wavelength']        

        xrf = XRSDFitter(q_I,pops,src_wl)
        fit_pops,rpt = xrf.fit(p_fix,p_b,p_c,errwtd,logIwtd,logqwtd)

        if self.message_callback:
            self.message_callback(xrf.print_report(pops,fit_pops,rpt))
 
        I_opt = compute_intensity(q_I[:,0],fit_pops,src_wl)

        from matplotlib import pyplot as plt
        I_guess = compute_intensity(q_I[:,0],pops,src_wl)
        plt.figure(2)
        plt.semilogy(q_I[:,0],q_I[:,1],'k')
        plt.semilogy(q_I[:,0],I_guess,'r')
        plt.semilogy(q_I[:,0],I_opt,'g')
        plt.show()

        q_I_opt = np.array([q_I[:,0],I_opt]).T
        self.outputs['populations'] = fit_pops 
        self.outputs['report'] = rpt
        self.outputs['q_I_opt'] = q_I_opt




import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools


# specify workflow names: 
wf_names = ['saxs_fit','main']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['main']['spectrum_files'] = 'IO.FILESYSTEM.BuildFileList'
op_maps['main']['batch'] = 'EXECUTION.Batch'
op_maps['main']['fit_params'] = 'PACKAGING.BATCH.XYDataFromBatch'

op_maps['saxs_fit']['read_spectrum'] = 'IO.NUMPY.Loadtxt_q_I_dI'
op_maps['saxs_fit']['fit_spectrum'] = 'PROCESSING.FITTING.XRSDFit'
op_maps['saxs_fit']['log_I'] = 'PROCESSING.BASIC.LogY'
op_maps['saxs_fit']['log_Ifit'] = 'PROCESSING.BASIC.LogY'
op_maps['saxs_fit']['output_CSV'] = 'IO.CSV.WriteArrayCSV'

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)

# define guinier-porod and noise populations
pops = {} 
pops['noise'] = dict(
    structure = 'diffuse',
    parameters = {'I0':0.1},
    basis = {'flat_noise':{'flat':{}}}
    )
pops['scatterers'] = dict(
    structure = 'diffuse',
    parameters = {},
    basis = {'gp_scatterers':{'guinier_porod':{'G':100.,'rg':10.,'D':4.}}}
    )

wf = wfmgr.workflows['saxs_fit']

# input 0: spectrum file
wf.set_op_input('read_spectrum','file_path',None) 
wf.connect_input('file_path','read_spectrum.inputs.file_path') 

# input 1: populations dict
wf.set_op_input('fit_spectrum','populations',pops)
wf.connect_input('populations','fit_spectrum.inputs.populations')

wf.connect_output('filename','read_spectrum.outputs.filename')
wf.connect_output('q_I','read_spectrum.outputs.q_I')
wf.connect_output('q_logI','log_I.outputs.x_logy')
wf.connect_output('q_I_opt','fit_spectrum.outputs.q_I_opt')
wf.connect_output('q_logI_opt','log_Ifit.outputs.x_logy')
wf.connect_output('fit_pops','fit_spectrum.outputs.populations')

wf.set_op_input('read_spectrum','delimiter',',')
wf.set_op_input('log_I','x_y','read_spectrum.outputs.q_I','workflow item')
wf.set_op_input('fit_spectrum','q_I','read_spectrum.outputs.q_I','workflow item')
wf.set_op_input('log_Ifit','x_y','fit_spectrum.outputs.q_I_opt','workflow item')
wf.set_op_input('output_CSV','array','fit_spectrum.outputs.q_I_opt','workflow item')
wf.set_op_input('output_CSV','headers',['q (1/angstrom)','intensity (arb)'])
wf.set_op_input('output_CSV','dir_path','read_spectrum.outputs.dir_path','workflow item')
wf.set_op_input('output_CSV','filename','read_spectrum.outputs.filename','workflow item')
wf.set_op_input('output_CSV','filetag','_fit')


wf = wfmgr.workflows['main']

# input 0: spectrum file location
wf.set_op_input('spectrum_files','dir_path',None) 
wf.connect_input('data_dir','spectrum_files.inputs.dir_path') 

# input 1: spectrum file regex
wf.set_op_input('spectrum_files','regex','*.csv') 
wf.connect_input('data_regex','spectrum_files.inputs.regex') 

wf.set_op_input('batch','work_item','saxs_fit','entire workflow')
wf.set_op_input('batch','input_arrays',['spectrum_files.outputs.file_list'],'workflow item')
wf.set_op_input('batch','input_keys',['file_path'])

wf.set_op_input('fit_params','batch_outputs','batch.outputs.batch_outputs','workflow item')
wf.set_op_input('fit_params','x_key','filename')
wf.set_op_input('fit_params','y_key','fit_params')


pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','SAXS','batch_saxs_fit.wfl'),wfmgr)




from collections import OrderedDict

from ...Operation import Operation

from xrsdkit.tools import piftools

inputs=OrderedDict(
    experiment_id=None,
    t_utc=None,
    temperature=None,
    q_I=None,
    populations=None)
outputs=OrderedDict(pif=None)

class PackXRSDPIF(Operation):
    """Build a PIF record for a SAXS spectrum"""

    def __init__(self):
        super(PackXRSDPIF,self).__init__(inputs,outputs)
        self.input_doc['experiment_id'] = 'string experiment id '\
            '(pif uid = experiment_id+"_"+t_utc)'
        self.input_doc['t_utc'] = 'time in seconds utc'
        self.input_doc['temperature'] = 'temperature of the system in degrees Celsius'
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding saxs intensities'
        self.input_doc['populations'] = 'dict enumerating scatterer populations'
        self.output_doc['pif'] = 'pif object representing the input data'

    def run(self):
        expt_id = self.inputs['experiment_id']
        t_utc = self.inputs['t_utc']
        pops = self.inputs['populations']
        uid_full = 'tmp'
        if expt_id is not None:
            uid_full = expt_id 
        if t_utc is not None:
            uid_full = uid_full+'_'+str(int(t_utc))
        temp_C = self.inputs['temperature']
        q_I = self.inputs['q_I']

        csys = piftools.make_pif(uid_full,expt_id,t_utc,q_I,temp_C,pops)

        self.outputs['pif'] = csys




import numpy as np
from collections import OrderedDict

import pyFAI

from ...Operation import Operation

inputs = OrderedDict(poni_dict=None) 
outputs = OrderedDict(integrator=None) 

class BuildPyFAIIntegrator(Operation):
    """Produce a PyFAI.AzimuthalIntegrator from a dict of PONI parameters.

    Input PONI dict should be similar 
    to the output of PyFAI.AzimuthalIntegrator.getPyFAI()
    """
    def __init__(self):
        super(BuildPyFAIIntegrator,self).__init__(inputs,outputs)
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_datatype['poni_dict'] = 'dict'
        self.output_doc['integrator'] = 'PyFAI.AzimuthalIntegrator object set up with input poni_dict'

    def run(self):
        pd = self.inputs['poni_dict']
        p = pyFAI.AzimuthalIntegrator()
        p.setPyFAI(**pd)
        self.outputs['integrator'] = p




import os
from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    col_headers=None,
    format_string=None,
    data_lists=None,
    dir_path=None,
    filename=None,
    filetag='')
outputs=OrderedDict(
    file_path=None,
    filename=None)

class WriteListsDAT(Operation):
    """Write a 2d array to a .dat file"""

    def __init__(self):
        super(WriteListsDAT, self).__init__(inputs, outputs)
        self.input_doc['col_headers'] = 'list of column headers (optional)'
        self.input_doc['format_string'] = 'list numpy format strings'
        self.input_doc['data_lists'] = 'list of lists of data- '\
            'there should be as many lists as there are `col_headers`'
        self.input_doc['dir_path'] = 'the path to the destination directory'
        self.input_doc['filename'] = 'the name of the file to be saved- no extension is expected'
        self.input_doc['filetag'] = 'tag appended to filename- no extension is expected'
        self.output_doc['file_path'] = 'the path to the finished dat file: dir_path+filename+filetag+.dat'
        self.output_doc['file_path'] = 'the name of the output file: filename+filetag'

    def run(self):
        col_hdrs = self.inputs['col_headers']
        p = self.inputs['dir_path']
        fnm = self.inputs['filename']
        tag = self.inputs['filetag']
        dat_path = os.path.join(p,self.inputs['filename']+tag+'.dat')
        self.outputs['file_path'] = dat_path
        self.outputs['filename'] = fnm+tag 
        dl = self.inputs['data_lists']

        ncols = len(dl)
        fmt = self.inputs['format_string']
        if fmt is None:
            fmt = ''
            for i in range(ncols):
                fmt += '{!s} \t'
            #if ncols > 0:
            #    fmt = fmt[:-2]

        outfile = open(dat_path,'w')

        n_data = len(dl[0])
        rows = [[col[ri] for col in dl] for ri in range(n_data)]
        #a = np.stack(dl).T

        #if fmt is None:
        #    fmt = '%.18e'
        if col_hdrs is not None:
            h_str = ''
            for hdr in col_hdrs:
                h_str += hdr+' \t'
            outfile.write(h_str + os.linesep)
        for r in rows:
            r_str = fmt.format(*r) + os.linesep
            outfile.write(r_str)
        outfile.close()
        #import pdb; pdb.set_trace()
        #np.savetxt(csv_path, a, fmt=fmt, delimiter=',', newline=os.linesep, header=h_str)




from collections import OrderedDict

import numpy as np

from ..Operation import Operation

inputs = OrderedDict(x_y=None)
outputs = OrderedDict(x=None,y=None)

class Unzip(Operation):
    """Unzip an array of x,y data"""

    def __init__(self):
        super(Unzip, self).__init__(inputs, outputs)
        self.output_doc['x_y'] = 'list of 2-element x,y iterables'
        self.input_doc['x'] = 'x values'
        self.input_doc['y'] = 'y values'

    def run(self):
        x_y = self.inputs['x_y']
        x = [itm[0] for itm in x_y]
        y = [itm[1] for itm in x_y]
        self.outputs['x'] = x
        self.outputs['y'] = y




from collections import OrderedDict
import copy
import time

from ..Operation import Operation
from .. import optools

inputs=OrderedDict(
    work_item=None,
    input_generators=[],
    input_keys=[],
    static_inputs=[],
    static_input_keys=[],
    delay=1.,
    max_delay=float('inf'),
    max_exec=float('inf'))

outputs=OrderedDict(
    realtime_inputs=None,
    realtime_outputs=None)

class Realtime(Operation):
    """Realtime-execute a Workflow or Operation"""

    def __init__(self):
        super(Realtime,self).__init__(inputs,outputs)
        self.input_doc['work_item'] = 'the Operation '\
            'or Workflow object to be executed in real time'

        self.input_doc['input_generators'] = 'one generator for each of '\
            'the `input_keys`, to generate inputs during execution- '\
            '`work_item` will be executed when all of the generators '\
            'return a value that is not None'
        self.input_doc['input_keys'] = 'list of keys for setting '\
            'inputs of the `work_item`- these should correspond to '\
            'either Operation.inputs or Workflow.inputs, '\
            'depending on whether `work_item` is an Operation or a Workflow'

        self.input_doc['static_inputs'] = 'one object for each of '\
            'the `work_item` inputs indicated by `static_input_keys`, '\
            'to be set to the same value for every execution'
        self.input_doc['static_input_keys'] = 'list of keys for setting '\
            'static inputs of the `work_item`- these should correspond to '\
            'either Operation.inputs or Workflow.inputs, '\
            'depending on whether `work_item` is an Operation or a Workflow'
            
        self.input_doc['delay'] = 'delay in seconds '\
            'between attempts to generate new inputs'
        self.input_doc['max_delay'] = '(optional) maximum delay'\
            'in seconds, before giving up and stopping execution'
        self.input_doc['max_exec'] = '(optional) maximum number '\
            'of executions of `work_item`'
            
        self.output_doc['realtime_inputs'] = 'list of dicts '\
            'containing [input_name:input_value] '\
            'for each execution of `work_item`'
        self.output_doc['realtime_outputs'] = 'list of dicts '\
            'containing [output_name:output_value] '\
            'for each execution of `work_item`'
        
    def run(self):
        wrk = self.inputs['work_item']
        inp_gens = self.inputs['input_generators']
        inpks = self.inputs['input_keys']
        stat_inps = self.inputs['static_inputs']
        stat_inpks = self.inputs['static_input_keys']
        maxdly = self.inputs['max_delay']
        maxexec = self.inputs['max_exec']
        dly = self.inputs['delay']

        self.outputs['realtime_inputs'] = [] 
        self.outputs['realtime_outputs'] = [] 
        if self.data_callback: 
            self.data_callback('outputs.realtime_inputs',[])
            self.data_callback('outputs.realtime_outputs',[])

        nx = 0 # total number of executions
        keep_going = True
        self.message_callback('STARTING REALTIME EXECUTION')
        while keep_going:
            # collect inputs from inp_gens
            inp_dict = OrderedDict.fromkeys(inpks)
            inps_ready = False
            nd = 0 # number of consecutive delays
            if nx >= maxexec:
                msg = 'Reached maximum executions ({}): Realtime stopping.'.format(maxexec) 
                self.message_callback(msg)
                keep_going = False
            while keep_going and not inps_ready:

                if self.stop_flag:
                    self.message_callback('Realtime execution stopped.')
                    return
                
                for k,gen in zip(inpks,inp_gens):
                    if inp_dict[k] is None:
                        inp_dict[k] = gen.next()
                if all([inp_dict[k] is not None for k in inpks]):
                    inps_ready = True
                    if self.data_callback: 
                        self.data_callback('outputs.realtime_inputs.'+str(nx),copy.deepcopy(inp_dict))
                else:
                    # delay
                    time.sleep(float(dly)) 
                    nd+=1
                    currentdly = nd*dly
                    #self.message_callback('... WAITING FOR INPUTS ({}/{} ms)'
                    #    .format(currentdly,maxdly))
                    if currentdly >= maxdly:
                        keep_going = False 
            if keep_going:
                for inpnm,inpval in inp_dict.items():
                    wrk.set_input(inpnm,inpval)
                if any(stat_inpks): 
                    for inpnm,inpval in zip(stat_inpks,stat_inps):
                        wrk.set_input(inpnm,inpval)
                self.message_callback('REALTIME RUN {}'.format(nx))
                wrk.run()
                out_dict = wrk.get_outputs()
                self.outputs['realtime_outputs'].append(out_dict)
                if self.data_callback: 
                    self.data_callback('outputs.realtime_outputs.'+str(nx),copy.deepcopy(out_dict))
                nx+=1
        self.message_callback('REALTIME EXECUTION STOPPED')




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import importlib

from ..models.TreeModel import TreeModel
from .. import plugins as pgns
from .. import pawstools
from .PawsPlugin import PawsPlugin

class PluginManager(TreeModel):
    """Tree for storing, browsing, and managing PawsPlugins"""

    def __init__(self):
        flag_dict = OrderedDict()
        flag_dict['select'] = False
        super(PluginManager,self).__init__(flag_dict)
        self.plugins = self._root_dict
        self.message_callback = self.tagged_print 
        # dict of clones for plugins across threads:
        #self.plugin_clones = OrderedDict()
        # dict of bools to keep track of who is at work:
        self.plugin_running = OrderedDict() 
        self.pool = None

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def add_plugin(self,plugin_name,plugin_module):
        """Import, name, and add a plugin.

        After a plugin is added to a plugin_manager,
        it is available as plugin_manager.plugins[plugin_name].

        Parameters
        ----------
        plugin_name : str
            Name that will be used to refer to this plugin after it is added.
        plugin_module : str
            Name of the plugin module.
            Example: If class MyPlugin is in the CATEGORY.MyPlugin module,
            retrieve it with `plugin_module` = 'CATEGORY.MyPlugin'.
        """
        p = self.get_plugin(plugin_module)
        if not self.is_tag_valid(plugin_name): 
            raise pawstools.PluginNameError(self.tag_error_message(plugin_name))
        #p.message_callback = self.message_callback
        p.data_callback = partial(self.set_plugin_item,plugin_name)
        self.set_item(plugin_name,p)
        self.plugin_running[plugin_name] = False 

    def set_plugin_item(self,pgn_name,item_uri,item_data):
        full_uri = pgn_name+'.'+item_uri
        self.set_item(full_uri,item_data)

    def get_plugin(self,plugin_module): 
        """Import, instantiate, return a PawsPlugin from its module.

        This can also be used to test the Python environment 
        for compatibility with a plugin.

        Parameters
        ----------
        plugin_module : str
            Name of the plugin module.
            See add_plugin().

        Returns
        -------
        PawsPlugin 
            An instance of the PawsPlugin subclass defined in `plugin_module`. 
        """
        mod = importlib.import_module('.'+plugin_module,pgns.__name__)
        return mod.__dict__[plugin_module]()

    def set_input(self,plugin_name,input_name,val,tp=None):
        """Set a plugin input to the provided value.

        Parameters
        ----------
        plugin_name : str
            Name that will be used to refer to this plugin after it is added.
        input_name : str
            name of the input to be set
        val : object
            the data to be used as plugin input
        tp : str or int, optional
            the input type determines how the PluginManager 
            interprets the input value `val`.
        """
        if not plugin_name in self.plugins.keys():
            raise KeyError('Plugin {} does not exist'.format(plugin_name))
        if not input_name in self.plugins[plugin_name].inputs.keys():
            raise KeyError('Input name {} not valid for PawsPlugin {} ({}).'
            .format(input_name,plugin_name,type(self.plugins[plugin_name]).__name__))
        if tp is not None and not tp in pawstools.valid_types and not tp in pawstools.input_types:
            # tp is neither a string or an enum
            raise ValueError('[{}] failed to parse input type: {}'.format(__name__,tp))
        il = self.plugins[plugin_name].input_locator[input_name]
        if tp is not None:
            if tp in pawstools.input_types:
                tp = pawstools.input_types.index(tp)
            il.tp = pawstools.valid_types[tp]
        if il.tp == pawstools.runtime_type:
            # set il.val as None to exclude this il.val from serialization
            il.val = None
        else:
            il.val = val
        if il.tp in [pawstools.basic_type,pawstools.runtime_type]:
            # these types may be loaded for immediate use
            self.set_item(plugin_name+'.inputs.'+input_name,val)

    def start_plugins(self,plugin_name_list):
        for pn in plugin_name_list:
            self.start_plugin(pn)

    def start_plugin(self,plugin_name):
        """Start the plugin referred to by `plugin_name`."""
        self.message_callback('starting plugin {}'.format(plugin_name))
        #pgn.plugin_clone = pgn.build_clone() 
        #self.plugin_clones[plugin_name] = pgn.plugin_clone
        self.prepare_plugin(plugin_name)
        self.plugin_running[plugin_name] = True
        #self.plugin_clones[plugin_name].start()
        self.plugins[plugin_name].start()

    def prepare_plugin(self,plugin_name):
        pgn = self.plugins[plugin_name]
        for inpname,il in pgn.input_locator.items():
            full_uri = plugin_name+'.inputs.'+inpname
            if il.tp == pawstools.plugin_item:
                if isinstance(il.val,list):
                    item_data = []
                    for item_name in il.val:
                        item = self.get_data_from_uri(item_name)
                        item_data.append(item)
                        #if item_name in self.plugins.keys():
                        #    if not self.plugin_running[item_name]:
                        #        self.start_plugin(item_name)
                else:
                    item_data = self.get_data_from_uri(il.val)
                    #if il.val in self.plugins.keys():
                    #    if not self.plugin_running[il.val]:
                    #        self.start_plugin(il.val)
            elif il.tp == pawstools.basic_type:
                item_data = il.val
            else:
                raise ValueError('Unable to set input type {} for plugin {}'
                .format(il.tp,plugin_name))
            self.set_item(full_uri,item_data)

    def stop_plugin(self,plugin_name):
        self.message_callback('stopping plugin {}'.format(plugin_name))
        #if plugin_name in self.plugin_clones.keys():
        #    self.plugin_clones[plugin_name].stop()
        self.plugins[plugin_name].stop()
        self.plugin_running[plugin_name] = False

    def load_plugin(self,plugin_name,plugin_setup_dict):
        """Load and set up a PawsPlugin, given its setup_dict().

        Parameters
        ----------
        plugin_name : str
            Name that will be used to refer to this plugin after it is added.
        plugin_setup_dict : dict 
            Dict specifying plugin setup,
            probably generated by calling the plugin's own setup_dict() method.
        """
        plugin_mod = plugin_setup_dict['plugin_module']
        self.add_plugin(plugin_name,plugin_mod)
        p = self.plugins[plugin_name]
        il_setup_dict = plugin_setup_dict['inputs']
        for inp_name in p.inputs.keys():
            if inp_name in il_setup_dict.keys():
                tp = il_setup_dict[inp_name]['tp']
                val = il_setup_dict[inp_name]['val']
                self.set_input(plugin_name,inp_name,val,tp)

    def n_plugins(self):
        """Return number of plugins currently loaded."""
        return len(self.plugins) 

    def build_tree(self,x):
        """Return a dict describing a tree-like structure of this object.

        This is a reimplemention of TreeModel.build_tree() 
        to define this object's child tree structure.
        For a PluginManager, a dict is provided for each PawsPlugin,
        where the dict contains the results of calling
        self.build_tree(plugin.inputs)
        """
        if isinstance(x,PawsPlugin):
            d = OrderedDict()
            d['inputs'] = self.build_tree(x.inputs)
            d.update(x.get_plugin_content())
        else:
            return super(PluginManager,self).build_tree(x) 
        return d





from __future__ import print_function
from collections import OrderedDict
import importlib

from ..models.TreeModel import TreeModel
from ..models.TreeItem import TreeItem
from .. import operations as ops
from ..operations.Operation import Operation

class OpManager(TreeModel):
    """OpManager provides access to and control over paws Operations."""

    def __init__(self):
        default_flags = OrderedDict()
        default_flags['active'] = False
        super(OpManager,self).__init__(default_flags)
        self.message_callback = self.tagged_print 
        self.cat_op_list = []
        self.load_operations()

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def get_operation(self,operation_uri):
        """Get an Operation, activate it if needed, instantiate, return.

        Parameters
        ----------
        operation_uri : str
            uri indicating the operation module,
            e.g. PROCESSING.TESTS.Fibonacci.

        Returns
        -------
        op : Operation
            the instantiated Operation 
        """
        if not self.is_op_activated(operation_uri):
            try:
                self.activate_op(operation_uri)
            except ImportError as ex:
                from paws.core.pawstools import OperationLoadError
                msg = 'Most likely, the system '\
                    'does not have the right dependencies '\
                    'for Operation {}'.format(operation_uri)
                raise OperationLoadError(msg) 
        op = self.get_data_from_uri(operation_uri)
        return op()

    def load_operations(self,cat_op_list=ops.cat_op_list):
        """Load Operations into OpManager.

        Parameters
        ----------
        cat_op_list : list of (str,str), optional
            Each element in the list is a tuple containing two strings.
            The first string in the tuple indicates the category subpackage.
            The second string in the tuple indicates the operation module.
            The default is to load all operations 
            detected by paws.core.operations at startup.
        """
        for cat_op in cat_op_list:
            self.add_cat(cat_op[0])
            if cat_op[1]:
                self.add_op(cat_op[0],cat_op[1])
                self.cat_op_list.append((cat_op[0],cat_op[1]))

    def add_cat(self,cat_module):
        """Add category `cat_module` to the tree."""
        itm = self._root_item
        cat_tags = cat_module.split('.')
        cat_uri = cat_tags[0]
        if not self.contains_uri(cat_uri):
            self.set_item(cat_uri,{})
        if len(cat_tags) > 1:
            for cat_tag in cat_tags[1:]:
                cat_uri = cat_uri+'.'+cat_tag
                if not self.contains_uri(cat_uri):
                    self.set_item(cat_uri,{})

    def add_op(self,cat_module,op_name):
        """Add label for an Operation at `op_name` under category `cat_module`."""
        op_uri = cat_module+'.'+op_name
        self.set_item(op_uri,None)

    def activate_op(self,op_module):
        """Import Operation module and add its Operation subclass to the tree.

        This method imports the Operation to check compatibility,
        and then sets the 'active' flag to True.
        After this, the Operation is available
        via self.get_op()

        Parameters
        ----------
        op_module : str
            Name of the Operation module.
            Example: If class MyOperation is in the CATEGORY.MyOperation module,
            retrieve it with `op_module` = 'CATEGORY.MyOperation'.
        """
        op_name = op_module.split('.')[-1]
        m = importlib.import_module('.'+op_module,ops.__name__)
        op = getattr(m,op_name)
        optest = op()
        self.set_item(op_module,op)
        op_itm = self.get_from_uri(op_module)
        self.set_flagged(op_itm,'active',True)

    def is_op_activated(self,op_module):
        """Return boolean indicating whether Operation is active.

        Parameters
        ----------
        op_module : str
            Name of the Operation module.
            see activate_op().
        """
        op_itm = self.get_from_uri(op_module)
        return op_itm.flags['active']

    def n_ops(self):
        return len(self.cat_op_list) 

    def list_operations(self):
        return [catnm+'.'+opnm for catnm,opnm in self.cat_op_list] 

    def print_cat(self,cat_uri,rowprefix='    '):
        """
        Generate a string that lists the contents 
        of the operations category specified by cat_uri
        """
        catdata = self.get_data_from_uri(cat_uri)
        tree_string = '\n' 
        if isinstance(catdata,dict):
            for k,x in catdata.items():
                if x is None:
                    # this should be the case for non-activated ops
                    tree_string = tree_string + rowprefix + '{} (disabled) \n'.format(k)
                elif isinstance(x,dict):
                    # this should be the case for a subcat
                    next_cat_tree = self.print_cat(cat_uri+'.'+k,rowprefix+'    ')
                    tree_string = tree_string + rowprefix + '{}: {}'.format(k,next_cat_tree)
                else:
                    # the only remaining case is an activated operation
                    tree_string = tree_string + rowprefix + '{} \n'.format(k)
        return tree_string





import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools

# specify workflow names: 
wf_names = ['background_process','sample_process','main']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['main']['read_calibration'] = 'IO.CALIBRATION.NikaToPONI'
op_maps['main']['build_integrator'] = 'PROCESSING.INTEGRATION.BuildPyFAIIntegrator'
op_maps['main']['background_files'] = 'IO.FILESYSTEM.BuildFileList'
op_maps['main']['background_batch'] = 'EXECUTION.Batch'
op_maps['main']['sample_files'] = 'IO.FILESYSTEM.FileIterator'
op_maps['main']['sample_realtime'] = 'EXECUTION.Realtime'

op_maps['background_process']['read_header'] = 'IO.BL15.ReadHeader_SSRL15'
op_maps['background_process']['time_temp'] = 'PACKAGING.BL15.TimeTempFromHeader'
op_maps['background_process']['image_path'] = 'IO.FILESYSTEM.BuildFilePath'
op_maps['background_process']['read_image'] = 'IO.IMAGE.FabIOOpen'
op_maps['background_process']['integrate'] = 'PROCESSING.INTEGRATION.ApplyIntegrator1d'
op_maps['background_process']['q_window'] = 'PACKAGING.Window'
op_maps['background_process']['dezinger'] = 'PROCESSING.ZINGERS.EasyZingers1d'
op_maps['background_process']['log_I'] = 'PROCESSING.BASIC.LogY'
op_maps['background_process']['output_CSV'] = 'IO.CSV.WriteArrayCSV'

op_maps['sample_process']['read_header'] = 'IO.BL15.ReadHeader_SSRL15'
op_maps['sample_process']['time_temp'] = 'PACKAGING.BL15.TimeTempFromHeader'
op_maps['sample_process']['image_path'] = 'IO.FILESYSTEM.BuildFilePath'
op_maps['sample_process']['read_image'] = 'IO.IMAGE.FabIOOpen'
op_maps['sample_process']['integrate'] = 'PROCESSING.INTEGRATION.ApplyIntegrator1d'
op_maps['sample_process']['q_window'] = 'PACKAGING.Window'
op_maps['sample_process']['dezinger'] = 'PROCESSING.ZINGERS.EasyZingers1d'
op_maps['sample_process']['bg_subtract'] = 'PROCESSING.BACKGROUND.BgSubtractByTemperature'
op_maps['sample_process']['log_I'] = 'PROCESSING.BASIC.LogY'
op_maps['sample_process']['output_CSV'] = 'IO.CSV.WriteArrayCSV'

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)



wf = wfmgr.workflows['main']

# input 0: path to calibration file 
wf.set_op_input('read_calibration','file_path',None)
wf.connect_input('nika_file','read_calibration.inputs.file_path')

# input 1: path to background header files directory
wf.set_op_input('background_files','dir_path',None)
wf.connect_input('background_dir','background_files.inputs.dir_path')

# input 2: regex for background header files 
wf.set_op_input('background_files','regex','*.txt')
wf.connect_input('background_regex','background_files.inputs.regex')

# input 3: path to sample header files directory
wf.set_op_input('sample_files','dir_path',None)
wf.connect_input('sample_dir','sample_files.inputs.dir_path')

# input 4: regex for sample header files 
wf.set_op_input('sample_files','regex','*.txt')
wf.connect_input('sample_regex','sample_files.inputs.regex')

# input 5: whether or not to process new files
wf.set_op_input('sample_files','new_files_only',False)
wf.connect_input('new_files_only','sample_files.inputs.new_files_only')

# input 6: maximum delay for realtime to bail 
wf.set_op_input('sample_realtime','max_delay',100000)
wf.connect_input('max_delay','sample_realtime.inputs.max_delay')

wf.set_op_input('build_integrator','poni_dict','read_calibration.outputs.poni_dict','workflow item')

wf.set_op_input('background_batch','work_item','background_process','entire workflow')
wf.set_op_input('background_batch','input_arrays',['background_files.outputs.file_list'],'workflow item')
wf.set_op_input('background_batch','input_keys',['header_file'])
wf.set_op_input('background_batch','static_inputs',['build_integrator.outputs.integrator'],'workflow item')
wf.set_op_input('background_batch','static_input_keys',['integrator'])

wf.set_op_input('sample_realtime','work_item','sample_process','entire workflow')
wf.set_op_input('sample_realtime','input_generators',['sample_files.outputs.file_iterator'],'workflow item')
wf.set_op_input('sample_realtime','input_keys',['header_file'])
wf.set_op_input('sample_realtime','static_inputs',
    ['build_integrator.outputs.integrator',
     'background_batch.outputs.batch_outputs'],
    'workflow item')
wf.set_op_input('sample_realtime','static_input_keys',
    ['integrator',
     'bg_batch_outputs'])




wf = wfmgr.workflows['background_process']

# input 0: path to header file
wf.set_op_input('read_header','file_path',None)
wf.connect_input('header_file','read_header.inputs.file_path')

# input 1: image directory, in case separate from headers
wf.set_op_input('image_path','dir_path',None)
wf.connect_input('image_dir','image_path.inputs.dir_path')

# inputs 2 and 3: q-range for data windowing
# TODO: pass in from batch
wf.set_op_input('q_window','x_min',0.0)
wf.set_op_input('q_window','x_max',1.0)
wf.connect_input('q_min','q_window.inputs.x_min')
wf.connect_input('q_max','q_window.inputs.x_max')

# input 4: key for fetching temperature from header dictionary 
wf.set_op_input('time_temp','temp_key','CTEMP')
wf.connect_input('temp_key','time_temp.inputs.temp_key')

# input 5: pyfai.AzimuthalIntegrator
# AzimthalIntegrators are not serializable,
# so specify this input as 'runtime' type
wf.set_op_input('integrate','integrator',None,'runtime')
wf.connect_input('integrator','integrate.inputs.integrator')

wf.connect_output('temperature','time_temp.outputs.temperature')
wf.connect_output('q_I','integrate.outputs.q_I')
wf.connect_output('q_I_dz','dezinger.outputs.q_I_dz')
wf.connect_output('q_logI_dz','log_I.outputs.x_logy')

wf.set_op_input('image_path','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('image_path','ext','tif')
wf.set_op_input('read_image','file_path','image_path.outputs.file_path','workflow item')
wf.set_op_input('time_temp','header_dict','read_header.outputs.header_dict','workflow item')
wf.set_op_input('time_temp','time_key','time')
wf.set_op_input('integrate','image_data','read_image.outputs.image_data','workflow item')
wf.set_op_input('q_window','x_y','integrate.outputs.q_I','workflow item')
wf.set_op_input('dezinger','q_I','q_window.outputs.x_y_window','workflow item')
wf.set_op_input('log_I','x_y','dezinger.outputs.q_I_dz','workflow item')
wf.set_op_input('output_CSV','array','dezinger.outputs.q_I_dz','workflow item')
wf.set_op_input('output_CSV','headers',['q (1/angstrom)','intensity (arb)'])
wf.set_op_input('output_CSV','dir_path','read_header.outputs.dir_path','workflow item')
wf.set_op_input('output_CSV','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('output_CSV','filetag','_dz')


wf = wfmgr.workflows['sample_process']

# input 0: path to header file
wf.set_op_input('read_header','file_path',None)
wf.connect_input('header_file','read_header.inputs.file_path')

# input 1: image directory, in case separate from headers
wf.set_op_input('image_path','dir_path',None)
wf.connect_input('image_dir','image_path.inputs.dir_path')

# inputs 2 and 3: q-range for data windowing
# TODO: pass in from batch
wf.set_op_input('q_window','x_min',0.0)
wf.set_op_input('q_window','x_max',1.0)
wf.connect_input('q_min','q_window.inputs.x_min')
wf.connect_input('q_max','q_window.inputs.x_max')

# input 4: key for fetching temperature from header dictionary 
wf.set_op_input('time_temp','temp_key','CTEMP')
wf.connect_input('temp_key','time_temp.inputs.temp_key')

# input 5: pyfai.AzimuthalIntegrator
# AzimthalIntegrators are not serializable,
# so specify this input as 'runtime' type
wf.set_op_input('integrate','integrator',None,'runtime')
wf.connect_input('integrator','integrate.inputs.integrator')

# input 6: the outputs of the background batch 
wf.connect_input('bg_batch_outputs','bg_subtract.inputs.bg_batch_outputs')
wf.set_op_input('bg_subtract','bg_batch_outputs',None)

wf.connect_output('time','time_temp.outputs.time')
wf.connect_output('temperature','time_temp.outputs.temperature')
wf.connect_output('q_I','integrate.outputs.q_I')
wf.connect_output('q_I_dz','dezinger.outputs.q_I_dz')
wf.connect_output('q_I_dz_bgsub','bg_subtract.outputs.q_I_bgsub')
wf.connect_output('q_logI_dz_bgsub','logI.outputs.x_logy')

wf.set_op_input('image_path','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('image_path','ext','tif')
wf.set_op_input('read_image','file_path','image_path.outputs.file_path','workflow item')
wf.set_op_input('time_temp','header_dict','read_header.outputs.header_dict','workflow item')
wf.set_op_input('time_temp','time_key','time')
wf.set_op_input('integrate','image_data','read_image.outputs.image_data','workflow item')
wf.set_op_input('q_window','x_y','integrate.outputs.q_I','workflow item')
wf.set_op_input('dezinger','q_I','q_window.outputs.x_y_window','workflow item')
wf.set_op_input('bg_subtract','q_I_meas','dezinger.outputs.q_I_dz','workflow item')
wf.set_op_input('bg_subtract','T_meas','time_temp.outputs.temperature','workflow item')
wf.set_op_input('bg_subtract','q_I_bg_key','q_I_dz')
wf.set_op_input('bg_subtract','T_bg_key','temperature')
wf.set_op_input('log_I','x_y','bg_subtract.outputs.q_I_bgsub','workflow item')
wf.set_op_input('output_CSV','array','bg_subtract.outputs.q_I_bgsub','workflow item')
wf.set_op_input('output_CSV','headers',['q (1/angstrom)','intensity (arb)'])
wf.set_op_input('output_CSV','filename','read_image.outputs.filename','workflow item')
wf.set_op_input('output_CSV','filetag','_dz_bgsub')

pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','SAXS','BL15','realtime_integrate_bgsub.wfl'),wfmgr)




from collections import OrderedDict
import glob
import os

from ...Operation import Operation

inputs=OrderedDict(
    dir_path=None,
    regex='*.tif')
outputs=OrderedDict(file_list=None)

class BuildFileList(Operation):
    """
    Read a directory and filter its contents with a regular expression
    to form a list of file paths.
    """

    def __init__(self):
        super(BuildFileList,self).__init__(inputs,outputs)
        self.input_doc['dir_path'] = 'path to directory containing files'
        self.input_doc['regex'] = 'regular expression to select files'
        
    def run(self):
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        fl = glob.glob(os.path.join(dirpath,rx))
        self.outputs['file_list'] = fl




from collections import OrderedDict

import numpy as np

from ...Operation import Operation
from xrsdkit.fitting.xrsd_fitter import XRSDFitter 

inputs = OrderedDict(
    q_I=None,
    populations=None,
    fixed_params=None,
    step_size=0.01,
    nsteps_burn=1000,
    nsteps_anneal=1000,
    nsteps_quench=1000,
    T_anneal=0.2,
    T_burn=0.3)
outputs = OrderedDict(
    best_params=None,
    final_params=None,
    report=None)

class SpectrumMCAnneal(Operation):
    """Anneal XRSD parameters by Metropolis-Hastings Monte Carlo.

    This Operation explores the space 
    of parameters for a XRSD pattern,
    seeking a globally optimal fit for some measured data.
    It is useful for refining optimizations 
    that tend to get stuck in local minima.
    """

    def __init__(self):
        super(SpectrumMCAnneal, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of measured data: '\
            'intensity (arb) with respect to scattering vector (1/Angstrom)'
        self.input_doc['populations'] = 'dict defining scatterer populations, xrsdkit format'
        self.input_doc['fixed_params'] = 'dict (subset of `params`) '\
            'indicating which `params` should be fixed during anneal'
        self.input_doc['step_size'] = 'random walk fractional step size'
        self.input_doc['nsteps_burn'] = 'number of iterations '\
            'to perform in the initial burn-off phase' 
        self.input_doc['nsteps_anneal'] = 'number of iterations '\
            'to perform in the annealing phase' 
        self.input_doc['nsteps_quench'] = 'number of iterations '\
            'to perform in the quenching (zero acceptance) phase' 
        self.output_doc['best_params'] = 'dict of scattering parameters '\
            'yielding the best fit over all trials'
        self.output_doc['final_params'] = 'dict of scattering parameters '\
            'obtained in the final step of the algorithm'
        self.output_doc['report'] = 'dict reporting '\
            'the number of steps and reject ratio'

    def run(self):
        q_I = self.inputs['q_I']
        pops = self.inputs['populations']
        par_fix = self.inputs['fixed_params']
        stepsz = self.inputs['step_size']
        ns_burn = self.inputs['nsteps_burn']
        ns_anneal = self.inputs['nsteps_anneal']
        ns_quench = self.inputs['nsteps_quench']
        T_burn = self.inputs['T_burn']
        T_anneal = self.inputs['T_anneal']

        p_init = pops
        p_best = pops
        p_fin = pops
        rpt = OrderedDict()
        xf = XRSDFitter(q_I,all_pops)
        # TODO: refactor this for new xrsdkit API

        # the burn phase is for escaping local minima.
        p_best,p_fin,rpt_burn = xf.MC_anneal_fit(stepsz,ns_burn,T_burn,par_fix)
        p_best_burn = p_best
        obj_init = rpt_burn['objective_init']
        obj_best = rpt_burn['objective_best']

        if self.message_callback:
            self.message_callback('finished burn. \ninit: {} \nbest: {} \nfinal: {} \nRR: {}'
            .format(rpt_burn['objective_init'],
                rpt_burn['objective_best'],
                rpt_burn['objective_final'],
                rpt_burn['reject_ratio']))

        # the anneal phase is expected to gravitate to regions of parameter space
        # where the fit objective is low.
        p_best,p_fin,rpt_anneal = xf.MC_anneal_fit(stepsz,ns_anneal,T_anneal,par_fix)
        if rpt_anneal['objective_best'] < obj_best:
            obj_best = rpt_anneal['objective_best']
        else:
            # if no better params were found during the anneal phase,
            # fall back on the best params from the burn phase
            p_best = p_best_burn
        if self.message_callback:
            self.message_callback('finished anneal. \ninit: {} \nbest: {} \nfinal: {} \nRR: {}'
            .format(rpt_anneal['objective_init'],
                rpt_anneal['objective_best'],
                rpt_anneal['objective_final'],
                rpt_anneal['reject_ratio']))

        # the quench phase is expected to stochastically descend, 
        # such that the objective never increases from one step to the next,
        # effectively sinking deeper into the current local minimum.
        p_best,p_fin,rpt_quench = xf.MC_anneal_fit(stepsz,ns_quench,0.,par_fix)
        if self.message_callback:
            self.message_callback('finished quench. \ninit: {} \nbest: {} \nfinal: {} \nRR: {}'
            .format(rpt_quench['objective_init'],
                rpt_quench['objective_best'],
                rpt_quench['objective_final'],
                rpt_quench['reject_ratio']))
        rpt = OrderedDict(
            objective_init = obj_init,
            objective_best = obj_best,
            objective_final = rpt_quench['objective_final'],
            reject_ratio = rpt_anneal['reject_ratio'])
        self.outputs['best_pops'] = p_best 
        self.outputs['final_pops'] = p_fin 
        self.outputs['report'] = rpt 




import os
from collections import OrderedDict
import re

import tifffile
import numpy as np

from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(
    image_data=None,
    image_header=None,
    dir_path=None,
    filename=None)

class ReadImageAndHeader_SSRL42(Operation):
    """Read image and header from SSRL beam line 4-2."""

    def __init__(self):
        super(ReadImageAndHeader_SSRL42, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to a tif file '\
            'produced by beamline 4-2 at SSRL. '\
            'A .prp header file is expected '\
            'in the same directory as this .tif file.'
        self.output_doc['image_data'] = 'the image data as an ndarray'
        self.output_doc['image_header'] = 'the header data as a python dictionary'
        self.output_doc['dir_path'] = 'the directory portion of the input file_path'
        self.output_doc['filename'] = 'the filename with path and extension stripped'

    def run(self):
        tif_path = self.inputs['file_path']
        dirpath,filename = os.path.split(tif_path)
        filename_noext = os.path.splitext(filename)[0]
        path_noext = os.path.splitext(tif_path)[0]
        hdr_file_path = path_noext + '.prp'
        self.outputs['dir_path'] = dirpath 
        self.outputs['filename'] = filename_noext 
        img = tifffile.imread(tif_path)
        # flip the image...
        self.outputs['image_data'] = np.array(img[::-1,:])
        d = OrderedDict()
        for l in open(hdr_file_path,'r').readlines():
            if not '=' in l:
                if l[:4] == 'Time':
                    d['time'] = l.split('Time this file was written: ')[1]
            else:
                kv = l.split('=')
                if kv[0] == 'Detector mode':
                    d[kv[0]] = kv[1].split(' ')[0].replace(';','')
                elif kv[0] in ['Beam energy','Pipe length',
                    'Horizontal position','Vertical position',
                    'dispx position','dispy position']:
                    # Split off and throw away the unit specification
                    d[kv[0]] = float(kv[1].split(' ')[0])
                elif kv[0] in ['Exposure time','Counting time','Phi position']:
                    # Units not specified: save float directly.
                    d[kv[0]] = float(kv[1])
                elif kv[0] in ['Scan motor','Scan range']:
                    d[kv[0]] = kv[1]
                else:
                    # Use key and value directly,
                    # use re.sub to filter out parentheticals 
                    d[re.sub('\(.*\)','',kv[0])] = float(re.sub('\(.*\)','',kv[1]))
            self.outputs['image_header'] = d




import numpy as np
from collections import OrderedDict

from ...Operation import Operation

inputs=OrderedDict(array_in=None)
outputs=OrderedDict(array_out=None)

class ArrayMirrorVertical(Operation):
    """Mirror an array across a vertical plane.
    
    Exchanges indices along axis 1.
    """

    def __init__(self):
        super(ArrayMirrorVertical, self).__init__(inputs, outputs)
        self.input_doc['array_in'] = '2d array'
        self.output_doc['array_out'] = 'input array mirrored vertically'

    def run(self):
        x = self.inputs['array_in']
        self.outputs['array_out'] = x[:,::-1]





from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs = OrderedDict(
    q=None,
    populations=None,
    params=None)
outputs = OrderedDict(
    I=None,
    q_I=None)

class ComputeXRSDPattern(Operation):
    """Compute an x-ray scattering and/or diffraction pattern"""

    def __init__(self):
        super(ComputeXRSDPattern, self).__init__(inputs, outputs)
        self.input_doc['q'] = 'array of wave vectors (1/Angstrom)'
        self.input_doc['populations'] = 'dict defining sample populations, in xrsdkit format'
        self.output_doc['I'] = 'array of computed intensities'
        self.output_doc['q_I'] = 'n-by-2 array of q and I'

    def run(self):
        q = self.inputs['q']
        pops = self.inputs['populations']
        pars = self.inputs['params']

        all_pops = OrderedDict.fromkeys(saxs_math.population_keys)
        all_pops.update(pops)
        I = saxs_math.compute_saxs(q,all_pops,pars)        

        self.outputs['I'] = I 
        self.outputs['q_I'] = np.vstack([q,I]).T 

